'''
Taken from the Instruction Induction paper: https://arxiv.org/pdf/2205.10782.pdf
Also: https://github.com/keirp/automatic_prompt_engineer/blob/main/experiments/evaluation/instruction_induction/utility.py
Adapted in this work.
'''

import re
import string
from collections import Counter

TASK_TO_METRIC = {'common_concept': 'f1', 'informal_to_formal': 'f1', 'orthography_starts_with': 'es', 'taxonomy_animal': 'es', 'synonyms': 'contains'}
default_metric = 'em'

def score_fn(prediction, ans, metric):
    if metric == 'f1':
        result = get_multi_answer_f1(prediction, ans)
    elif metric == 'es':
        result = get_multi_answer_exact_set(prediction, ans)
    elif metric == 'contains':
        result = get_multi_answer_contains(prediction, ans)
    elif metric == 'em':
        result = get_multi_answer_em(prediction, ans)
    return result

def normalize_prediction(prediction, lowercase=True):
    prediction = prediction.replace(' and ', ' ')
    prediction = prediction.replace('Sentence 1:', ' ')
    prediction = prediction.replace('Sentence 2:', ' ')
    prediction = prediction.strip()
    prediction = prediction.split("\n")[0]
    prediction = prediction.split(".")[0]

    if lowercase:
        prediction = prediction.lower()

    # remove punctuation
    prediction = prediction.replace('-', ' ')
    prediction = prediction.translate(
        str.maketrans('', '', string.punctuation))

    return prediction


def get_f1_score(prediction, ground_truth):
    prediction_tokens = normalize_prediction(
        prediction, lowercase=True).split()
    ground_truth_tokens = normalize_prediction(
        ground_truth, lowercase=True).split()
    common = Counter(prediction_tokens) & Counter(ground_truth_tokens)
    num_same = sum(common.values())
    if num_same == 0:
        return 0
    precision = 1.0 * num_same / len(prediction_tokens)
    recall = 1.0 * num_same / len(ground_truth_tokens)
    f1 = (2 * precision * recall) / (precision + recall)
    return f1


def get_em_score(prediction, ground_truth):
    prediction_normalized = normalize_prediction(prediction, lowercase=True)
    ground_truth_normalized = normalize_prediction(
        ground_truth, lowercase=True)
    return prediction_normalized == ground_truth_normalized


def get_exact_set_score(prediction, ground_truth):
    prediction_normalized = normalize_prediction(
        prediction, lowercase=True).split()
    ground_truth_normalized = normalize_prediction(
        ground_truth, lowercase=True).split()
    return int(set(prediction_normalized) == set(ground_truth_normalized))


def get_contains_score(prediction, ground_truth):
    prediction_normalized = normalize_prediction(prediction, lowercase=True)
    ground_truth_normalized = normalize_prediction(
        ground_truth, lowercase=True)
    if re.search(r'\b({0})\b'.format(ground_truth_normalized), prediction_normalized):
        return 1


def get_multi_answer_em(prediction, answers):
    for answer in answers:
        if get_em_score(prediction, answer) == 1:
            return 1
    return 0


def get_multi_answer_f1(prediction, answers):
    f1_scores = []
    for answer in answers:
        f1_scores.append(get_f1_score(prediction, answer))
    return max(f1_scores)


def get_multi_answer_exact_set(prediction, answers):
    for answer in answers:
        if get_exact_set_score(prediction, answer) == 1:
            return 1
    return 0


def get_multi_answer_contains(prediction, answers):
    for answer in answers:
        if get_contains_score(prediction, answer) == 1:
            return 1
    return 0


class AbstractTrainer:
    def __init__(self, args, logger):
        self.args = args
        self.logger = logger

    def initialize(self, task):
        # initialize from (1) files or (2) induction from few examples
        raise NotImplementedError

    def train(self, model, initial_node, task):
        # returns a best prompt and the final metric on the train set
        raise NotImplementedError
    
    def evaluate(self, model, node, task, split="test"):
        # returns a result_df and a final performance score
        raise NotImplementedError
        
    def load_session(self, path):
        # load checkpoints for resuming from crashes
        raise NotImplementedError
        
    def save_session(self, path):
        # save session at each timestamp
        raise NotImplementedError


import sys
sys.path.append("../src")
import json

from greaterprompt.optimizer import ApeOptimizer, GreaterDataloader

import openai
import streamlit as st


with st.sidebar:
    openai_key = st.markdown("<h3>🔐 OpenAI Key</h3>", unsafe_allow_html=True)
    openai_key = st.text_input("OpenAI Key", type="password")
    settings = st.markdown("<h3>🛠️ Advanced Settings</h3>", unsafe_allow_html=True)
    task_model = st.selectbox("Task Model", ["openai_gpt35_turbo_instruct", "openai_gpt4", "openai_gpt4_turbo", "openai_gpt4o", "openai_gpt4o_mini"], index=0)
    optim_model = st.selectbox("Optim Model", ["openai_gpt35_turbo_instruct", "openai_gpt4", "openai_gpt4_turbo", "openai_gpt4o", "openai_gpt4o_mini"], index=2)

st.markdown("<h1 style='text-align: center; white-space: nowrap;'>🤩 Optimize with APE Optimizer</h1>", unsafe_allow_html=True)

uploaded_file = st.file_uploader("Upload a jsonl input file", type=("jsonl"))
p_init = st.text_input("P Initial")
col1, col2, col3 = st.columns([1, 2, 1])
with col2:
    start_button = st.button("⚡ Start Optimization", type="primary", use_container_width=True)

if not openai_key and start_button:
    st.info("Please enter an OpenAI key to start")

if not uploaded_file and start_button:
    st.info("Please upload a jsonl file to start")

if p_init.strip() == "" and start_button:
    st.info("Please enter a valid P Initial to start")

if openai_key and uploaded_file and p_init.strip() and start_button:
    openai.api_key = openai_key
    
    content = uploaded_file.getvalue().decode("utf-8")
    lines = content.strip().split('\n')

    inputs = []
    
    for line in lines:
        if line.strip():
            try:
                inputs.append(json.loads(line))
            except json.JSONDecodeError:
                st.error(f"Invalid JSON in line: {line}")

    optimize_config = {
        "task_model": task_model,
        "optim_model": optim_model,
    }

    with st.status("⚡ Optimizing...", expanded=True) as status:
        ape_optimizer = ApeOptimizer(optimize_config=optimize_config)
        dataset = GreaterDataloader(custom_inputs=inputs)

        output = ape_optimizer.optimize(dataset, p_init)

        status.update(label="😄 Optimization complete!", state="complete")
        st.markdown("<h5 style='text-align: left; white-space: nowrap;'>🎯 Optimization Result:</h5>", unsafe_allow_html=True)
        st.markdown(f"<p>{output}</p>", unsafe_allow_html=True)

        if output:
            st.download_button(
                label="📥 Download Optimized Result",
                data=json.dumps(output, indent=2, ensure_ascii=False),
                file_name="optimized_prompt.json",
                mime="application/json"
            )



from setuptools import setup, find_packages

with open("requirements.txt") as f:
    requirements = f.read().splitlines()

with open("README.md") as readme_file:
    readme = readme_file.read()

setup(
    name="textgrad",
    version="0.1.4",
    description="",
    python_requires=">=3.9",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.8",
    ],
    license="MIT license",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/zou-group/textgrad",
    author="Zou Group",
    author_email="merty@stanford.edu",
    packages=find_packages(include=["textgrad", "textgrad.*"]),
    include_package_data=True,
    install_requires=requirements,
    extras_require={
        "vllm": ["vllm"],
    },
    zip_safe=False,
)




from .engine import EngineLM, get_engine
from .variable import Variable
from typing import List, Union
from .autograd import LLMCall, FormattedLLMCall, OrderedFieldsMultimodalLLMCall
from .autograd import Module
from .config import SingletonBackwardEngine


class TextLoss(Module):
    def __init__(self, 
                 eval_system_prompt: Union[Variable, str],
                 engine: Union[EngineLM, str] = None):
        """
        A vanilla loss function to evaluate a response.
        In particular, this module is used to evaluate any given text object.

        :param evaluation_instruction: The evaluation instruction variable.
        :type evaluation_instruction: Variable
        :param engine: The EngineLM object.
        :type engine: EngineLM
        
        :example:
        >>> from textgrad import get_engine, Variable
        >>> from textgrad.loss import TextLoss
        >>> engine = get_engine("gpt-4o")
        >>> evaluation_instruction = Variable("Is ths a good joke?", requires_grad=False)
        >>> response_evaluator = TextLoss(evaluation_instruction, engine)
        >>> response = Variable("What did the fish say when it hit the wall? Dam.", requires_grad=True)
        >>> response_evaluator(response)
        """
        super().__init__()
        if isinstance(eval_system_prompt, str):
            eval_system_prompt = Variable(eval_system_prompt, requires_grad=False, role_description="system prompt for the evaluation")
        self.eval_system_prompt = eval_system_prompt
        if ((engine is None) and (SingletonBackwardEngine().get_engine() is None)):
            raise Exception("No engine provided. Either provide an engine as the argument to this call, or use `textgrad.set_backward_engine(engine)` to set the backward engine.")
        elif engine is None:
            engine = SingletonBackwardEngine().get_engine()
        if isinstance(engine, str):
            engine = get_engine(engine)
        self.engine = engine
        self.llm_call = LLMCall(self.engine, self.eval_system_prompt)

    def forward(self, instance: Variable):
        """
        Calls the ResponseEvaluation object.

        :param instance: The instance variable.
        :type instance: Variable
        :return: The result of the evaluation
        """
        return self.llm_call(instance)

class MultiFieldEvaluation(Module):
    def __init__(
        self,
        evaluation_instruction: Variable,
        role_descriptions: List[str],
        engine: Union[EngineLM, str] = None,
        system_prompt: Variable = None,
    ):
        """A module to compare two variables using a language model.

        :param evaluation_instruction: Instruction to use as prefix for the comparison, specifying the nature of the comparison.
        :type evaluation_instruction: Variable
        :param engine: The language model to use for the comparison.
        :type engine: EngineLM
        :param v1_role_description: Role description for the first variable, defaults to "prediction to evaluate"
        :type v1_role_description: str, optional
        :param v2_role_description: Role description for the second variable, defaults to "correct result"
        :type v2_role_description: str, optional
        :param system_prompt: System prompt to use for the comparison, defaults to "You are an evaluation system that compares two variables."
        :type system_prompt: Variable, optional
        
        :example:
        TODO: Add an example
        """
        super().__init__()
        self.evaluation_instruction = evaluation_instruction
        if ((engine is None) and (SingletonBackwardEngine().get_engine() is None)):
            raise Exception("No engine provided. Either provide an engine as the argument to this call, or use `textgrad.set_backward_engine(engine)` to set the backward engine.")
        elif engine is None:
            engine = SingletonBackwardEngine().get_engine()
        if isinstance(engine, str):
            engine = get_engine(engine)
        self.engine = engine
        self.role_descriptions = role_descriptions
        if system_prompt:
            self.system_prompt = system_prompt
        else:
            self.system_prompt = Variable("You are an evaluation system that compares two variables.",
                                            requires_grad=False,
                                            role_description="system prompt for the evaluation")
        format_string_items = ["{{instruction}}"]
        for role_description in role_descriptions:
            format_string_items.append(f"**{role_description}**: {{{role_description}}}")
        format_string = "\n".join(format_string_items)
        self.format_string = format_string.format(instruction=self.evaluation_instruction, **{role_description: "{"+role_description+"}" for role_description in role_descriptions})
        self.fields = {"instruction": self.evaluation_instruction, **{role_description: None for role_description in role_descriptions}}
        self.formatted_llm_call = FormattedLLMCall(engine=self.engine,
                                                   format_string=self.format_string,
                                                   fields=self.fields,
                                                   system_prompt=self.system_prompt)

    def forward(self, inputs: List[Variable]):
        for role_description, var in zip(self.role_descriptions, inputs):
            var.set_role_description(role_description)
        inputs_call = {"instruction": self.evaluation_instruction, 
                       **{role_description: var for role_description, var in zip(self.role_descriptions, inputs)}}
        return self.formatted_llm_call(inputs=inputs_call,
                                       response_role_description=f"evaluation of the a prediction")


class MultiFieldTokenParsedEvaluation(MultiFieldEvaluation):
    def __init__(
        self,
        evaluation_instruction: Variable,
        role_descriptions: List[str],
        engine: Union[EngineLM, str] = None,
        system_prompt: Variable = None,
        parse_tags: List[str] = None,
    ):
        super().__init__(
            evaluation_instruction=evaluation_instruction,
            role_descriptions=role_descriptions,
            engine=engine,
            system_prompt=system_prompt,
        )
        self.parse_tags = parse_tags

    def parse_output(self, response: Variable) -> str:
        """
        Parses the output response and returns the parsed response.

        :param response: The response to be parsed.
        :type response: Variable
        :return: The parsed response.
        :rtype: str
        """
        response_text = response.value
        response = response_text.split(self.parse_tags[0])[1].split(self.parse_tags[1])[0].strip()
        return response


DEFAULT_TEST_TIME = "You are an intelligent assistant used as an evaluator, and part of an optimization system. You will analyze a solution to a multi-choice problem. Investigate the reasoning and answer. Do not try to solve the problem, only raise the potential issues and mistakes in the answer. Be creative, think about different perspectives, and be very critical."

class MultiChoiceTestTime(Module):
    def __init__(self,
                 evaluation_instruction: str,
                 engine: Union[EngineLM, str] = None,
                 system_prompt: Variable = None):
        """
        The test-time loss to use when working on a response to a multiple choice question.

        :param evaluation_instruction: Instruction to guide the test time evaluation. This will be a prefix to the prompt.
        :type evaluation_instruction: str
        :param engine: LLM engine to use for the test-time loss computation.
        :type engine: EngineLM
        :param system_prompt: System prompt for the test-time loss computation, defaults to None
        :type system_prompt: Variable, optional
        """
        super().__init__()
        if system_prompt:
            self.tt_system_prompt = system_prompt
        else:
            tt_system_prompt = DEFAULT_TEST_TIME
            self.tt_system_prompt = Variable(tt_system_prompt,
                                                requires_grad=False,
                                                role_description="system prompt for the test-time evaluation")
        
        if ((engine is None) and (SingletonBackwardEngine().get_engine() is None)):
            raise Exception("No engine provided. Either provide an engine as the argument to this call, or use `textgrad.set_backward_engine(engine)` to set the backward engine.")
        elif engine is None:
            engine = SingletonBackwardEngine().get_engine()
        if isinstance(engine, str):
            engine = get_engine(engine)
        self.engine = engine
        format_string = "{instruction}\nQuestion: {{question}}\nAnswer by the language model: {{prediction}}"
        self.format_string = format_string.format(instruction=evaluation_instruction)
        self.fields = {"prediction": None, "question": None}
        self.formatted_llm_call = FormattedLLMCall(engine=self.engine,
                                                   format_string=self.format_string,
                                                   fields=self.fields,
                                                   system_prompt=self.tt_system_prompt)

    def forward(self, question: str, prediction: Variable) -> Variable:
        question_variable = Variable(question, 
                                     requires_grad=False, 
                                     role_description="the multiple choice question")

        inputs = {"prediction": prediction, "question": question_variable}
        return self.formatted_llm_call(inputs=inputs,
                                       response_role_description=f"evaluation of the {prediction.get_role_description()}")

class ImageQALoss(Module):
    def __init__(self,
                 evaluation_instruction: str,
                 engine: Union[EngineLM, str] = None,
                 system_prompt: Variable = None):
        super().__init__()
        self.evaluation_instruction = Variable(evaluation_instruction, role_description="evaluation instruction", requires_grad=False)
        if ((engine is None) and (SingletonBackwardEngine().get_engine() is None)):
            raise Exception("No engine provided. Either provide an engine as the argument to this call, or use `textgrad.set_backward_engine(engine)` to set the backward engine.")
        elif engine is None:
            engine = SingletonBackwardEngine().get_engine()
        if isinstance(engine, str):
            engine = get_engine(engine)
        self.engine = engine
        if system_prompt:
            self.system_prompt = system_prompt
        else:
            self.system_prompt = Variable("You are an evaluation system that evaluates image-related questions.",
                                            requires_grad=False,
                                            role_description="system prompt for the evaluation")

        self.multimodal_llm_call = OrderedFieldsMultimodalLLMCall(engine=self.engine,
                                                                  system_prompt=self.system_prompt,
                                                                  fields=["Evaluation Instruction", "Question", "Image", "Answer"])

    def forward(self, image: Variable, question: Variable, response: Variable) -> Variable:
        
        inputs = {
            "Evaluation Instruction": self.evaluation_instruction,
            "Question": question,
            "Image": image,
            "Answer": response
        }
        return self.multimodal_llm_call(inputs=inputs,
                                        response_role_description=f"evaluation of the {response.get_role_description()}")


import random
import os
import guidance 

from .default_trainer import DefaultTrainer
from .utils import clean_string, _load_prompt
from .node import Node

class PE2Trainer(DefaultTrainer):
    def prepare_meta_prompts(self):

        # initializer
        init_prompt = _load_prompt(os.path.join(self.args.meta_prompts_dir, "initializer.md"))
        generation_config = "n={} temperature={} max_tokens={}".format(self.args.init_n_prompts, self.args.init_temperature, self.args.prompt_max_tokens)
        prompt = init_prompt.replace("[[GENERATION_CONFIG]]", generation_config)
        self.initializer = guidance(prompt, llm=self.llm) # guidance program for initialization
        
        # proposer: ~ optimizer.step()
        proposer_prompt = _load_prompt(os.path.join(self.args.meta_prompts_dir, "proposer.md"))
        self.proposer = guidance(proposer_prompt, llm=self.llm)
        
        if self.args.optim_use_instruction:
            assert "{{instruction}}" in proposer_prompt
            self.instruction_prompt = _load_prompt(os.path.join(self.args.meta_prompts_dir, "instruction.md"))
        if self.args.optim_use_demonstrations:
            assert "{{demonstrations}}" in proposer_prompt
            self.demonstrations_prompt = _load_prompt(os.path.join(self.args.meta_prompts_dir, "demonstrations.md"))
        if self.args.optim_use_optim_tutorial:
            assert "{{optim_tutorial}}" in proposer_prompt
            self.optim_tutorial_prompt = _load_prompt(os.path.join(self.args.meta_prompts_dir, "optim_tutorial.md"))
        
    def _update(self, node, batch, optim_batch):
        examples = self._pack_batch(batch)

        guidance.llms.OpenAI.cache.clear()
        full_prompt = self.model.full_prompt.replace("{{input}}", "<input>").replace("{{instruction}}", node.prompt)
        # prepare input args
        proposer_args = {
            "prompt": node.prompt, "timestamp": node.timestamp, 
            "full_prompt": full_prompt, 
            "step_size": None,
            # placeholder with None so that guidance won't complain
            "batch_size": None, "failures": None,
            "history": None, "instruction": None, "demonstrations": None, "optim_tutorial": None,
            "max_tokens": self.args.prompt_max_tokens
        }
        
        proposer_args["batch_size"] = self.args.batch_size
        proposer_args["examples"] = examples
        if self.args.optim_use_step_size:
            proposer_args["step_size"] = self.args.step_size
        if self.args.optim_use_momentum:
            proposer_args["history"] = node.history
        if self.args.optim_use_instruction:
            proposer_args["instruction"] = self.instruction_prompt
        if self.args.optim_use_demonstrations:
            proposer_args["demonstrations"] = self.demonstrations_prompt
        if self.args.optim_use_optim_tutorial:
            proposer_args["optim_tutorial"] = self.optim_tutorial_prompt

        # guidance program
        f = self.proposer(**proposer_args)

        new_prompt = clean_string(f["new_prompt"])
        
        # post-process optimization history
        if self.args.optim_use_momentum:
            new_history = clean_string(f["new_history"])
            full_history = node.history + "\n" + new_history if node.timestamp > 0 else new_history
        else:
            full_history = None
            
        reasoning = clean_string(f["reasoning"]) + "[SEP]" + clean_string(f["reasoning2"]) if "reasoning2" in f else clean_string(f["reasoning"])
        new_node = Node(
            timestamp=node.timestamp+1,
            id=node.id+"-{}".format(node.n_child),
            parent=node.id,
            prompt=new_prompt,
            history=full_history,
            optim_batch=optim_batch,
            reasoning=reasoning
        )

        node.n_child += 1

        return new_node


import argparse
import concurrent
#from dotenv import load_dotenv
import sys
import os

parent_parent_folder = os.path.abspath(os.path.join(os.getcwd(), '../..'))

local_textgrad_folder = os.path.join(parent_parent_folder, 'textgrad_ollm')

sys.path.insert(0, local_textgrad_folder)

print(sys.path)

from textgrad.engine.local_model_openai_api import ChatExternalClient
import textgrad as tg

#load_dotenv(override=True)

from tqdm import tqdm
from textgrad.tasks import load_task

import numpy as np
import random

def set_seed(seed):
    np.random.seed(seed)
    random.seed(seed)

def config():
    parser = argparse.ArgumentParser(description="Optimize a prompt for a task.")
    parser.add_argument("--task", type=str, default="BBH_object_counting", help="The task to evaluate the model on.")
    parser.add_argument("--evaluation_engine", type=str, default="meta-llama/Llama-2-7b-chat-hf", help="The API to use for evaluation.")
    parser.add_argument("--test_engine", type=str, default="meta-llama/Llama-2-7b-chat-hf", help="The API to use for evaluation.")
    parser.add_argument("--batch_size", type=int, default=3, help="The batch size to use for training.")
    parser.add_argument("--max_epochs", type=int, default=3, help="The maximum number of epochs to train for.")
    parser.add_argument("--seed", type=int, default=42)
    parser.add_argument("--run_validation", action="store_true", help="Whether to run validation or not.")
    parser.add_argument("--do_not_run_larger_model", action="store_true", help="Whether to run the larger model or not.")
    parser.add_argument("--num_threads", type=int, default=1, help="The number of threads to use for evaluation.")
    return parser.parse_args()

args = config()

def eval_sample(item, eval_fn, model):
    x, y = item
    y = str(y)
    x = tg.Variable(x, requires_grad=False, role_description="query to the language model")
    y = tg.Variable(str(y), requires_grad=False, role_description="correct answer for the query")
    response = model(x)
    try:
        eval_output_variable = eval_fn(inputs=dict(prediction=response, ground_truth_answer=y))
        return int(eval_output_variable.value)
    except:
        eval_output_variable = eval_fn([x, y, response])
        eval_output_parsed = eval_fn.parse_output(eval_output_variable)
        return int(eval_output_parsed)
    

def eval_dataset(test_set, eval_fn, model, max_samples: int=None):
    if max_samples is None:
        max_samples = len(test_set)
    accuracy_list = []
    with concurrent.futures.ThreadPoolExecutor(max_workers=args.num_threads) as executor:
        futures = []
        for _, sample in enumerate(test_set):
            
            future = executor.submit(eval_sample, sample, eval_fn, model)
            futures.append(future)
            if len(futures) >= max_samples:
                break
        tqdm_loader = tqdm(concurrent.futures.as_completed(futures), total=len(futures), position=0)
        for future in tqdm_loader:
            acc_item = future.result()
            accuracy_list.append(acc_item)
            tqdm_loader.set_description(f"Proxy Accuracy: {np.mean(accuracy_list)}") # Actual accuracy requires very thorough two stage extraction
    return accuracy_list 

def run_validation_revert(system_prompt: tg.Variable, results, model, eval_fn, val_set):
    val_performance = np.mean(eval_dataset(val_set, eval_fn, model))
    previous_performance = np.mean(results["validation_acc"][-1])
    print("val_performance: ", val_performance)
    print("previous_performance: ", previous_performance)
    previous_prompt = results["prompt"][-1]
    
    if val_performance < previous_performance:
        print(f"rejected prompt: {system_prompt.value}")
        system_prompt.set_value(previous_prompt)
        val_performance = previous_performance

    results["validation_acc"].append(val_performance)


set_seed(args.seed)

evaluation_engine = ChatExternalClient(client=None, model_string=args.evaluation_engine)

if args.evaluation_engine != args.test_engine:
    test_engine = ChatExternalClient(client=None, model_string=args.test_engine)
else:
    test_engine = evaluation_engine

llm_api_eval = evaluation_engine
llm_api_test = test_engine

# llm_api_eval = tg.get_engine(engine_name=args.evaluation_engine)
# llm_api_test = tg.get_engine(engine_name=args.test_engine)
tg.set_backward_engine(llm_api_eval, override=True)

# Load the data and the evaluation function
train_set, val_set, test_set, eval_fn = load_task(args.task, evaluation_api=llm_api_eval)
print("Train/Val/Test Set Lengths: ", len(train_set), len(val_set), len(test_set))
STARTING_SYSTEM_PROMPT = train_set.get_task_description()

train_loader = tg.tasks.DataLoader(train_set, batch_size=args.batch_size, shuffle=True)
print(STARTING_SYSTEM_PROMPT)


# Testing the 0-shot performance of the evaluation engine
system_prompt = tg.Variable(STARTING_SYSTEM_PROMPT, 
                            requires_grad=True, 
                            role_description="system prompt to the language model")
model_evaluation = tg.BlackboxLLM(llm_api_eval, system_prompt)

if not args.do_not_run_larger_model:
    reference = np.mean(eval_dataset(test_set, eval_fn, model_evaluation))


system_prompt = tg.Variable(STARTING_SYSTEM_PROMPT, 
                            requires_grad=True,
                            role_description="structured system prompt to a somewhat capable language model that specifies the behavior and strategies for the QA task")
model = tg.BlackboxLLM(llm_api_test, system_prompt)

optimizer = tg.TextualGradientDescent(engine=llm_api_eval, parameters=[system_prompt])

results = {"test_acc": [], "prompt": [], "validation_acc": []}
results["test_acc"].append(eval_dataset(test_set, eval_fn, model))
results["validation_acc"].append(eval_dataset(val_set, eval_fn, model))
results["prompt"].append(system_prompt.get_value())


for epoch in range(args.max_epochs):
    for steps, (batch_x, batch_y) in enumerate((pbar := tqdm(train_loader, position=0))):
        pbar.set_description(f"Training step {steps}. Epoch {epoch}")
        optimizer.zero_grad()
        losses = []
        for (x, y) in zip(batch_x, batch_y):
            y = str(y)
            x = tg.Variable(x, requires_grad=False, role_description="query to the language model")
            y = tg.Variable(y, requires_grad=False, role_description="correct answer for the query")
            response = model(x)
            try:
                eval_output_variable = eval_fn(inputs=dict(prediction=response, ground_truth_answer=y))
            except:
                eval_output_variable = eval_fn([x, y, response])
            losses.append(eval_output_variable)
        total_loss = tg.sum(losses)
        total_loss.backward()
        optimizer.step()
        if args.run_validation:
            run_validation_revert(system_prompt, results, model, eval_fn, val_set)
        print("sys prompt: ", system_prompt)
        test_acc = eval_dataset(test_set, eval_fn, model)
        results["test_acc"].append(test_acc)
        results["prompt"].append(system_prompt.get_value())
        if steps == 3:
            break

# Also dump the final results
import json

os.makedirs("./figures", exist_ok=True)
str_test_engine = args.test_engine.replace("/", "_")
str_evaluation_engine = args.evaluation_engine.replace("/", "_")
with open(f"./figures/results_{args.task}_{str_test_engine}_{str_evaluation_engine}.json", "w") as f:
    json.dump(results, f)


CODE_INSTANCE_ROLE_DESCRIPTION = "Code generated that must be evaluated for correctness and runtime performance"
SYSTEM_PROMPT_FOR_FIRST_CODE = """You are an AI that only responds with python code, NOT ENGLISH. You will be given a function signature and its docstring by the user. Write your full implementation (restate the function signature).
Use a Python code block to write your response. For example:
```python
print('Hello world!')
```"""

from textgrad.autograd.function import Module
from textgrad.autograd import FormattedLLMCall
from textgrad import EngineLM
from textgrad import Variable

DEFAULT_TEST_TIME_WITH_TESTS = """You are an intelligent assistant used as an evaluator, and part of an optimization system. 
You will analyze a code implementation for a coding problem and unit test results. 
The code will be tested with harder tests, so do not just check if the code passes the provided tests.
Think about the correctness of the code and its performance in harder test cases.
Give very concise feedback.
Investigate the code problem and the provided implementation. 
For each failed unit test case, start analyzing it by saying "The code did not pass this test because...". 
Explain why the current implementation is not getting the expected output. Do not provide a revised implementation. Carefully 
suggest why there are issues with the code and provide feedback.
"""

default_instruction_test = (f"")

class CodeTestTimewithTests(Module):
    def __init__(self,
                 engine: EngineLM,
                 evaluation_instruction: str = default_instruction_test,
                 system_prompt: Variable = None):
        super().__init__()
        if system_prompt:
            self.tt_system_prompt = system_prompt
        else:
            tt_system_prompt = DEFAULT_TEST_TIME_WITH_TESTS
            self.tt_system_prompt = Variable(tt_system_prompt,
                                                requires_grad=False,
                                                role_description="system prompt for the evaluation of the code solution")
        self.engine = engine
        format_string = "You are a language model that evaluates a python code snippet.\n"
        format_string += "{instruction}\n**The coding problem:**\n\n{{problem}}\n**{role}**{{program}}**\n\nThe test results:**\n\n{{tests}}\n"
        self.format_string = format_string.format(instruction=evaluation_instruction, role=CODE_INSTANCE_ROLE_DESCRIPTION)
        self.fields = {"problem": None, "program": None, "tests": None}
        self.formatted_llm_call = FormattedLLMCall(engine=self.engine,
                                                   format_string=self.format_string,
                                                   fields=self.fields,
                                                   system_prompt=self.tt_system_prompt)

    def forward(self, problem: str, program: Variable, tests: str) -> Variable:
        problem_variable = Variable(problem,
                                     requires_grad=False,
                                     role_description="the coding problem")
        tests = Variable(tests,
                         requires_grad=False,
                         role_description="the results of the unit tests on the code solution")
        inputs = {"program": program, "problem": problem_variable, "tests": tests}
        return self.formatted_llm_call(inputs=inputs,
                                       response_role_description=f"evaluation of the {program.get_role_description()}")



import os
import re
import tqdm
import retrying

import openai
from vllm import LLM, SamplingParams

from .model import AbstractModel
from .utils import create_batches, get_answer_cleansing_prompt

FULL_PRMOPT = """Q: {{input}}
A: {{instruction}}
"""

class ZeroshotCoTModel(AbstractModel):
    def __init__(self, args, logger):
        super().__init__(args, logger)
        if any(model in self.args.task_model for model in ["Mistral-7B-Instruct-v0.2", "mpt-7b-instruct", "Yi-6B"]):
        # if self.args.task_model in ["mistralai/Mistral-7B-Instruct-v0.2", "mosaicml/mpt-7b-instruct", "01-ai/Yi-6B"]:
            logger.info("Setting up `{}` model in vLLM".format(self.args.task_model))
            self.llm = LLM(model=self.args.task_model)

    def load_prompt(self):
        self.full_prompt = FULL_PRMOPT
    
    # @retrying.retry(
    #     stop_max_attempt_number=100, # Maximum number of attempts
    #     wait_exponential_multiplier=1000, wait_exponential_max=10000, 
    #     # Wait 2^x * 1000 milliseconds between each retry, up to 10 seconds, then 10 seconds afterwards
    #     retry_on_exception=lambda e: isinstance(e, Exception)
    # )
    def _request_instruct(self, queries):
        OPENAI_INSTRUCT_MODELS = {"openai_td003": "text-davinci-003", "openai_gpt35_turbo_instruct": "gpt-3.5-turbo-instruct"}
        engine = OPENAI_INSTRUCT_MODELS[self.args.task_model]

        responses = openai.Completion.create(
            engine=engine,
            prompt=queries,
            max_tokens=400,
            temperature=0.0,
            n=1,
            top_p=1,
            stream=False,
            stop=None
        )
        
        assert len(responses.choices) == len(queries)
        
        all_reasoning = [choice.text.strip() for choice in responses.choices]

        new_queries = []
        for query, choice in zip(queries, responses.choices):
            new_query = query + choice.text.strip() + " \n" + get_answer_cleansing_prompt(self.args)
            new_queries.append(new_query)

        responses = openai.Completion.create(
            engine=engine,
            prompt=new_queries,
            max_tokens=100 if self.args.subtask == "word_sorting" else 10, # word sorting answer is significantly longer
            temperature=0.0,
            n=1,
            top_p=1,
            stream=False,
            stop=None
        )

        assert len(responses.choices) == len(queries)

        all_raw_output = [choice.text.strip() for choice in responses.choices]

        return all_reasoning, all_raw_output

    @retrying.retry(
        stop_max_attempt_number=100, # Maximum number of attempts
        wait_exponential_multiplier=1000, wait_exponential_max=10000, 
        # Wait 2^x * 1000 milliseconds between each retry, up to 10 seconds, then 10 seconds afterwards
        retry_on_exception=lambda e: isinstance(e, Exception)
    )
    def _request_chat(self, queries):
        OPENAI_CHAT_MODELS = {"openai_gpt35_turbo": "gpt-3.5-turbo-1106", "openai_gpt4_turbo": "gpt-4-0125-preview"}
        engine = OPENAI_CHAT_MODELS[self.args.task_model]

        # client = OpenAI()
        all_reasoning, all_raw_output = [], []
        for query in queries:
            response = openai.ChatCompletion.create(
                model=engine,
                messages=[
                    {"role": "system", "content": "You are a helpful assistant."},
                    {"role": "user", "content": query},
                ],
                max_tokens=400,
                temperature=0.0,
                n=1,
                top_p=1,
                stream=False,
                stop=None
            )
            reasoning = response.choices[0].message.content
            all_reasoning.append(reasoning)

            # Second API call to do answer cleansing
            response = openai.ChatCompletion.create(
                model=engine,
                messages=[
                    {"role": "system", "content": "You are a helpful assistant."},
                    {"role": "user", "content": query},
                    {"role": "assistant", "content": reasoning},
                    {"role": "user", "content": get_answer_cleansing_prompt(self.args)},
                ],
                max_tokens=10,
                temperature=0.0,
                n=1,
                top_p=1,
                stream=False,
                stop=None
            )
            final_answer = response.choices[0].message.content
            all_raw_output.append(final_answer)

        return all_reasoning, all_raw_output
    
    def _request_vllm(self, queries):
        all_reasoning, all_raw_output, new_queries = [], [], []

        # first call
        sampling_params = SamplingParams(temperature=0.0, 
                                         max_tokens=400, n=1,
                                         top_p=1, stop=None)
        responses = self.llm.generate(queries, sampling_params)

        for query, response in zip(queries, responses):
            reasoning = response.outputs[0].text
            all_reasoning.append(reasoning)
            new_query = query + reasoning + " \n" + get_answer_cleansing_prompt(self.args)
            new_queries.append(new_query)

        # second call
        responses = self.llm.generate(new_queries, sampling_params)
        max_tokens = 100 if self.args.subtask == "word_sorting" else 10
        sampling_params = SamplingParams(temperature=0.0, 
                                         max_tokens=max_tokens, n=1,
                                         top_p=1, stop=None)
        responses = self.llm.generate(new_queries, sampling_params)

        for response in responses:
            final_answer = response.outputs[0].text
            all_raw_output.append(final_answer)
        
        return all_reasoning, all_raw_output


    def _request(self, queries):
        if self.args.task_model in ["openai_td003", "openai_gpt35_turbo_instruct"]:
            return self._request_instruct(queries) # legacy openai completion endpoints
        elif self.args.task_model in ["openai_gpt35_turbo", "openai_gpt4_turbo"]:
            return self._request_chat(queries) # chat endpoints which may needs additional formating
        elif any(model in self.args.task_model for model in ["Mistral-7B-Instruct-v0.2", "mpt-7b-instruct", "Yi-6B"]):
            return self._request_vllm(queries)
        else:
            raise NotImplementedError

    def _run_batch(self, instruction, batch):
        all_inputs = batch["input"].tolist()
        all_queries = []
        
        prompt = self.full_prompt.replace("{{instruction}}", instruction)
        
        for i in range(len(all_inputs)):
            query = prompt.replace("{{input}}", all_inputs[i])
            all_queries.append(query)
                               
        # response = self._request_substrate(all_queries)
        # all_raw_output = [choice["text"] for choice in response["choices"]]
        all_reasoning, all_raw_output = self._request(all_queries)
        
        return all_reasoning, all_raw_output
    
    def run(self, instruction, data):

        # if self.args.task_model in ["mistralai/Mistral-7B-Instruct-v0.2", "mosaicml/mpt-7b-instruct", "01-ai/Yi-6B"]:
        #     openai.default_headers = {"x-foo": "true"}
        #     openai.api_key = "EMPTY"
        #     openai.api_base = "http://localhost:8001/v1"

        data = data.copy()
        batch_size = 10 if self.args.task_model in ["openai_td003", "openai_gpt35_turbo_instruct"] else len(data)
        batches = create_batches(data, batch_size=batch_size)
        
        all_raw_output = []
        all_reasoning = []
        for batch in tqdm.tqdm(batches, desc="Batches"):
            batch_reasoning, batch_raw_output = self._run_batch(instruction, batch)
            all_reasoning += batch_reasoning
            all_raw_output += batch_raw_output

        data["reasoning"] = all_reasoning
        data["raw_output"] = all_raw_output

        return data
    
    def run_till_enough_errors(self, instruction, data, task, target_n_errors):
        data = data.copy().sample(frac=1.0, random_state=42) # random shuffle
        batch_size = 10 if self.args.task_model in ["openai_td003", "openai_gpt35_turbo_instruct"] else 20
        batches = create_batches(data, batch_size=batch_size)
        
        all_raw_output = []
        all_reasoning = []
        n_done = 0

        for batch in tqdm.tqdm(batches, desc="Batches"):
            batch_reasoning, batch_raw_output = self._run_batch(instruction, batch)
            
            n_done += len(batch)
            all_reasoning += batch_reasoning
            all_raw_output += batch_raw_output

            data_done = data.head(n_done).copy()
            data_done["reasoning"] = all_reasoning
            data_done["raw_output"] = all_raw_output

            data_done, _ = task.evaluate(data_done)
            n_errors = (data_done['score'] == 0.0).sum()
            if n_errors >= target_n_errors:
                self.logger.info("Found {} model mistakes by running {} examples".format(n_errors, n_done))
                break

        return data_done



from typing import Union
from .variable import Variable
from .autograd import LLMCall
from .autograd.function import Module
from .engine import EngineLM, get_engine
from .config import SingletonBackwardEngine

class BlackboxLLM(Module):
    def __init__(self, engine: Union[EngineLM, str] = None, system_prompt: Union[Variable, str] = None):
        """
        Initialize the LLM module.

        :param engine: The language model engine to use.
        :type engine: EngineLM
        :param system_prompt: The system prompt variable, defaults to None.
        :type system_prompt: Variable, optional
        """
        if ((engine is None) and (SingletonBackwardEngine().get_engine() is None)):
            raise Exception("No engine provided. Either provide an engine as the argument to this call, or use `textgrad.set_backward_engine(engine)` to set the backward engine.")
        elif engine is None:
            engine = SingletonBackwardEngine().get_engine()
        if isinstance(engine, str):
            engine = get_engine(engine)
        self.engine = engine
        if isinstance(system_prompt, str):
            system_prompt = Variable(system_prompt, requires_grad=False, role_description="system prompt for the language model")
        self.system_prompt = system_prompt
        self.llm_call = LLMCall(self.engine, self.system_prompt)

    def parameters(self):
        """
        Get the parameters of the blackbox LLM.

        :return: A list of parameters.
        :rtype: list
        """
        params = []
        if self.system_prompt:
            params.append(self.system_prompt)
        return params

    def forward(self, x: Variable) -> Variable:
        """
        Perform an LLM call.

        :param x: The input variable.
        :type x: Variable
        :return: The output variable.
        :rtype: Variable
        """
        return self.llm_call(x)




# from dotenv import load_dotenv
# from utils import get_output_of_test, timeout_exec
#
# class PythonEvaluator:
#     """
#     This is a simple evaluator for Python code.
#     """
#
#     def __init__(self):
#         raise Exception("You should remove this and remove the comments, but run at your own risk!!!"
#                         "This is going to run model-generated code on your machine.")
#
#     def evaluate(self, code, tests):
#         """
#         This function evaluates the code with the given tests.
#         Returns a string that contains the tests that passed and failed.
#
#
#         :param code:
#         :param tests:
#         :return:
#         """
#         raise Exception("You should remove this and remove the comments, but run at your own risk!!!"
#                         "This is going to run model-generated code on your machine.")
#         func_test_list = [f'from typing import *\n{code}\n\n{test}' for test in tests]
#
#         success_tests = []
#         failed_tests = []
#         num_tests = len(func_test_list)
#         for i in range(num_tests):
#
#             try:
#                 #timeout_exec(func_test_list[i])
#                 success_tests += [tests[i]]
#             except Exception as e:
#                 failed_test = tests[i]
#
#                 try:
#                     #output = get_output_of_test(code, failed_test)
#                     asserted_value = tests[i].split("==")[1].strip()
#                     failed_tests += [f"{tests[i]} # ERROR: This unit test fails. Output was {output}, but expected value was: {asserted_value}"]
#                 except Exception as e:
#                     failed_tests += [f"{tests[i]} # ERROR: This unit test fails because the function generated: {e}."]
#
#         state = []
#         for test in tests:
#             if test in success_tests:
#                 state += [True]
#             else:
#                 state += [False]
#
#         state = tuple(state)
#
#         feedback = "**Tests that the code passed:**\n"
#         if len(success_tests) == 0:
#             feedback += "\nNo tests passed.\n"
#         else:
#             for test in success_tests:
#                 feedback += f"\n{test}"
#         feedback += "\n\n**Tests that the code failed:**\n"
#         if len(failed_tests) == 0:
#             feedback += "\nNo tests failed.\n"
#         else:
#             for test in failed_tests:
#                 feedback += f"\n{test}"
#
#         return state, feedback


from .. import logger
from ..defaults import (SYSTEM_PROMPT_DEFAULT_ROLE, 
                               VARIABLE_OUTPUT_DEFAULT_ROLE)
from ..variable import Variable
from ..engine import EngineLM
from ..config import validate_engine_or_get_default
from typing import List, Dict
from .llm_backward_prompts import (
    EVALUATE_VARIABLE_INSTRUCTION,
    CONVERSATION_START_INSTRUCTION_BASE,
    CONVERSATION_START_INSTRUCTION_CHAIN,
    CONVERSATION_TEMPLATE,
    OBJECTIVE_INSTRUCTION_CHAIN,
    OBJECTIVE_INSTRUCTION_BASE,
    BACKWARD_SYSTEM_PROMPT,
    IN_CONTEXT_EXAMPLE_PROMPT_ADDITION,
)
from .function import Function, BackwardContext


class LLMCall(Function):
    def __init__(self, engine: EngineLM, system_prompt: Variable = None):
        """The simple LLM call function. This function will call the LLM with the input and return the response, also register the grad_fn for backpropagation.

        :param engine: engine to use for the LLM call
        :type engine: EngineLM
        :param system_prompt: system prompt to use for the LLM call, default depends on the engine.
        :type system_prompt: Variable, optional
        """
        super().__init__()
        self.engine = validate_engine_or_get_default(engine)
        self.system_prompt = system_prompt
        if self.system_prompt and self.system_prompt.get_role_description() is None:
            self.system_prompt.set_role_description(SYSTEM_PROMPT_DEFAULT_ROLE)
    
    def forward(self, input_variable: Variable, response_role_description: str = VARIABLE_OUTPUT_DEFAULT_ROLE) -> Variable:
        """
        The LLM call. This function will call the LLM with the input and return the response, also register the grad_fn for backpropagation.
        
        :param input_variable: The input variable (aka prompt) to use for the LLM call.
        :type input_variable: Variable
        :param response_role_description: Role description for the LLM response, defaults to VARIABLE_OUTPUT_DEFAULT_ROLE
        :type response_role_description: str, optional
        :return: response sampled from the LLM
        :rtype: Variable
        
        :example:
        >>> from textgrad import Variable, get_engine
        >>> from textgrad.autograd.llm_ops import LLMCall
        >>> engine = get_engine("gpt-3.5-turbo")
        >>> llm_call = LLMCall(engine)
        >>> prompt = Variable("What is the capital of France?", role_description="prompt to the LM")
        >>> response = llm_call(prompt, engine=engine) 
        # This returns something like Variable(data=The capital of France is Paris., grads=)
        """
        # TODO: Should we allow default roles? It will make things less performant.
        system_prompt_value = self.system_prompt.value if self.system_prompt else None

        # Make the LLM Call
        response_text = self.engine(input_variable.value, system_prompt=system_prompt_value)

        # Create the response variable
        response = Variable(
            value=response_text,
            predecessors=[self.system_prompt, input_variable] if self.system_prompt else [input_variable],
            role_description=response_role_description
        )
        
        logger.info(f"LLMCall function forward", extra={"text": f"System:{system_prompt_value}\nQuery: {input_variable.value}\nResponse: {response_text}"})
        
        # Populate the gradient function, using a container to store the backward function and the context
        response.set_grad_fn(BackwardContext(backward_fn=self.backward, 
                                             response=response, 
                                             prompt=input_variable.value, 
                                             system_prompt=system_prompt_value))

        return response
    
    def backward(self, response: Variable, prompt: str, system_prompt: str, backward_engine: EngineLM):
        """
        Backward pass through the LLM call. This will register gradients in place.

        :param response: The response variable.
        :type response: Variable
        :param prompt: The prompt string that will be used as input to an LM.
        :type prompt: str
        :param system_prompt: The system prompt string.
        :type system_prompt: str
        :param backward_engine: The backward engine that will do the gradient computation.
        :type backward_engine: EngineLM

        :return: None
        """
        children_variables = response.predecessors
        if response.get_gradient_text() == "":
            self._backward_through_llm_base(children_variables, response, prompt, system_prompt, backward_engine)
        else:
            self._backward_through_llm_chain(children_variables, response, prompt, system_prompt, backward_engine)

    @staticmethod
    def _construct_llm_chain_backward_prompt(backward_info: Dict[str, str]) -> str:
        conversation = CONVERSATION_TEMPLATE.format(**backward_info)
        backward_prompt = CONVERSATION_START_INSTRUCTION_CHAIN.format(conversation=conversation, **backward_info)
        backward_prompt += OBJECTIVE_INSTRUCTION_CHAIN.format(**backward_info)
        backward_prompt += EVALUATE_VARIABLE_INSTRUCTION.format(**backward_info)
        return backward_prompt

    @staticmethod
    def _backward_through_llm_chain(variables: List[Variable], 
                                    response: Variable, 
                                    prompt: str, 
                                    system_prompt: str,
                                    backward_engine: EngineLM):

        """
        Backward through the LLM to compute gradients for each variable, in the case where the output has gradients on them.
        i.e. applying the chain rule.
        
        :param variables: The list of variables to compute gradients for.
        :type variables: List[Variable]
        :param response: The response variable.
        :type response: Variable
        :param prompt: The prompt string.
        :type prompt: str
        :param system_prompt: The system prompt string.
        :type system_prompt: str
        :param backward_engine: The backward engine to use for computing gradients.
        :type backward_engine: EngineLM

        :return: None
        """
        for variable in variables:
            if not variable.requires_grad:
                continue

            backward_info = {
                "response_desc": response.get_role_description(),
                "response_value": response.get_value(),
                "response_gradient": response.get_gradient_text(),
                "prompt": prompt,
                "system_prompt": system_prompt,
                "variable_desc": variable.get_role_description(),
                "variable_short": variable.get_short_value()
            }
            
            backward_prompt = LLMCall._construct_llm_chain_backward_prompt(backward_info)

            logger.info(f"_backward_through_llm prompt", extra={"_backward_through_llm": backward_prompt})
            gradient_value = backward_engine(backward_prompt, system_prompt=BACKWARD_SYSTEM_PROMPT)
            logger.info(f"_backward_through_llm gradient", extra={"_backward_through_llm": gradient_value})
            
            var_gradients = Variable(value=gradient_value, role_description=f"feedback to {variable.get_role_description()}")
            variable.gradients.add(var_gradients)
            conversation = CONVERSATION_TEMPLATE.format(**backward_info)
            variable.gradients_context[var_gradients] = {
                "context": conversation, 
                "response_desc": response.get_role_description(),
                "variable_desc": variable.get_role_description()
            }
            
            if response._reduce_meta:
                var_gradients._reduce_meta.extend(response._reduce_meta)
                variable._reduce_meta.extend(response._reduce_meta)

    @staticmethod
    def _construct_llm_base_backward_prompt(backward_info: Dict[str, str]) -> str:
        conversation = CONVERSATION_TEMPLATE.format(**backward_info)
        backward_prompt = CONVERSATION_START_INSTRUCTION_BASE.format(conversation=conversation, **backward_info)
        backward_prompt += OBJECTIVE_INSTRUCTION_BASE.format(**backward_info)
        backward_prompt += EVALUATE_VARIABLE_INSTRUCTION.format(**backward_info)
        return backward_prompt

    @staticmethod
    def _backward_through_llm_base(variables: List[Variable], 
                                   response: Variable,
                                   prompt: str,
                                   system_prompt: str,
                                   backward_engine: EngineLM):
        """
        Backward pass through the LLM base. 
        In this case we do not have gradients on the output variable.

        :param variables: A list of variables to compute gradients for.
        :type variables: List[Variable]
        :param response: The response variable.
        :type response: Variable
        :param prompt: The prompt string.
        :type prompt: str
        :param system_prompt: The system prompt string.
        :type system_prompt: str
        :param backward_engine: The backward engine to use for computing gradients.
        :type backward_engine: EngineLM
        """
        for variable in variables:
            if not variable.requires_grad:
                continue

            backward_info = {
                "response_desc": response.get_role_description(),
                "response_value": response.get_value(),
                "prompt": prompt,
                "system_prompt": system_prompt,
                "variable_desc": variable.get_role_description(),
                "variable_short": variable.get_short_value()
            }
            
            backward_prompt = LLMCall._construct_llm_base_backward_prompt(backward_info)
            
            logger.info(f"_backward_through_llm prompt", extra={"_backward_through_llm": backward_prompt})
            gradient_value = backward_engine(backward_prompt, system_prompt=BACKWARD_SYSTEM_PROMPT)
            logger.info(f"_backward_through_llm gradient", extra={"_backward_through_llm": gradient_value})

            conversation = CONVERSATION_TEMPLATE.format(**backward_info)
            var_gradients = Variable(value=gradient_value, role_description=f"feedback to {variable.get_role_description()}")
            variable.gradients.add(var_gradients)
            variable.gradients_context[var_gradients] = {
                "context": conversation, 
                "response_desc": response.get_role_description(),
                "variable_desc": variable.get_role_description()
            }

            if response._reduce_meta:
                var_gradients._reduce_meta.extend(response._reduce_meta)
                variable._reduce_meta.extend(response._reduce_meta)



class FormattedLLMCall(LLMCall):
    def __init__(self, 
                 engine: EngineLM, 
                 format_string: str,
                 fields: Dict[str, str],
                 system_prompt: Variable = None):
        """This class is responsible for handling the formatting of the input before calling the LLM.
        It inherits from the LLMCall class and reuses its backward function.

        :param engine: The engine to use for the LLM call.
        :type engine: EngineLM
        :param format_string: The format string to use for the input. For instance, "The capital of {country} is {capital}". For a format string like this, we'll expect to have the fields dictionary to have the keys "country" and "capital". Similarly, in the forward pass, we'll expect the input variables to have the keys "country" and "capital".
        :type format_string: str
        :param fields: The fields to use for the format string. For the above example, this would be {"country": {}, "capital": {}}. This is currently a dictionary in case we'd want to inject more information later on.
        :type fields: dict[str, str]
        :param system_prompt: The system prompt to use for the LLM call. Default value depends on the engine.
        :type system_prompt: Variable, optional
        """
        super().__init__(engine, system_prompt)
        self.format_string = format_string
        self.fields = fields
    
    
    def forward(self, 
                inputs: Dict[str, Variable],
                response_role_description: str = VARIABLE_OUTPUT_DEFAULT_ROLE) -> Variable:
        """The LLM call with formatted strings. 
        This function will call the LLM with the input and return the response, also register the grad_fn for backpropagation.

        :param inputs: Variables to use for the input. This should be a mapping of the fields to the variables.
        :type inputs: dict[str, Variable]
        :param response_role_description: Role description for the response variable, defaults to VARIABLE_OUTPUT_DEFAULT_ROLE
        :type response_role_description: str, optional
        :return: Sampled response from the LLM
        :rtype: Variable
        """
        # First ensure that all keys are present in the fields
        assert set(inputs.keys()) == set(self.fields.keys()), f"Expected fields {self.fields.keys()} but got {inputs.keys()}"
        
        input_variables = list(inputs.values())
        
        # Now format the string
        formatted_input_string = self.format_string.format(**{k: inputs[k].value for k in inputs.keys()})

        # TODO: Should we allow default roles? It will make things less performant.
        system_prompt_value = self.system_prompt.value if self.system_prompt else None

        # Make the LLM Call
        response_text = self.engine(formatted_input_string, system_prompt=system_prompt_value)

        # Create the response variable
        response = Variable(
            value=response_text,
            predecessors=[self.system_prompt, *input_variables] if self.system_prompt else [*input_variables],
            role_description=response_role_description
        )
        
        logger.info(f"LLMCall function forward", extra={"text": f"System:{system_prompt_value}\nQuery: {formatted_input_string}\nResponse: {response_text}"})
        
        # Populate the gradient function, using a container to store the backward function and the context
        response.set_grad_fn(BackwardContext(backward_fn=self.backward, 
                                             response=response, 
                                             prompt=formatted_input_string, 
                                             system_prompt=system_prompt_value))
        
        return response


class LLMCall_with_in_context_examples(LLMCall):
    
    def forward(self, input_variable: Variable, response_role_description: str = VARIABLE_OUTPUT_DEFAULT_ROLE, in_context_examples: List[str]=None) -> Variable:
        """
        The LLM call. This function will call the LLM with the input and return the response, also register the grad_fn for backpropagation.
        
        :param input_variable: The input variable (aka prompt) to use for the LLM call.
        :type input_variable: Variable
        :param response_role_description: Role description for the LLM response, defaults to VARIABLE_OUTPUT_DEFAULT_ROLE
        :type response_role_description: str, optional
        :return: response sampled from the LLM
        :rtype: Variable
        
        :example:
        >>> from textgrad import Variable, get_engine
        >>> from textgrad.autograd.llm_ops import LLMCall
        >>> engine = get_engine("gpt-3.5-turbo")
        >>> llm_call = LLMCall(engine)
        >>> prompt = Variable("What is the capital of France?", role_description="prompt to the LM")
        >>> response = llm_call(prompt, engine=engine) 
        # This returns something like Variable(data=The capital of France is Paris., grads=)
        """
        # TODO: Should we allow default roles? It will make things less performant.
        system_prompt_value = self.system_prompt.value if self.system_prompt else None

        # Make the LLM Call
        response_text = self.engine(input_variable.value, system_prompt=system_prompt_value)


        # Create the response variable
        response = Variable(
            value=response_text.split('<FINAL>')[1].split('</FINAL>')[0].strip(),
            predecessors=[self.system_prompt, input_variable] if self.system_prompt else [input_variable],
            role_description=response_role_description
        )


        
        logger.info(f"LLMCall function forward", extra={"text": f"System:{system_prompt_value}\nQuery: {input_variable.value}\nResponse: {response_text}"})
        
        # Populate the gradient function, using a container to store the backward function and the context
        response.set_grad_fn(BackwardContext(backward_fn=self.backward, 
                                             response=response, 
                                             prompt=input_variable.value, 
                                             system_prompt=system_prompt_value,
                                             in_context_examples=in_context_examples))
        
        # Stopping creteria Could be improved, now is very ugly. 
        if "The plan doesn't need to be improved" in response_text.split('<FINAL>')[1].split('</FINAL>')[0].strip():
            response = None

        return response

    def backward(self, response: Variable, prompt: str, system_prompt: str, in_context_examples: List[str], backward_engine: EngineLM):
        """
        Backward pass through the LLM call. This will register gradients in place.

        :param response: The response variable.
        :type response: Variable
        :param prompt: The prompt string that will be used as input to an LM.
        :type prompt: str
        :param system_prompt: The system prompt string.
        :type system_prompt: str
        :param backward_engine: The backward engine that will do the gradient computation.
        :type backward_engine: EngineLM

        :return: None
        """
        children_variables = response.predecessors

        if response.get_gradient_text() == "":
            self._backward_through_llm_base(children_variables, response, prompt, system_prompt, backward_engine, in_context_examples)
        else:
            self._backward_through_llm_chain(children_variables, response, prompt, system_prompt, backward_engine, in_context_examples)

    @staticmethod
    def _construct_llm_chain_backward_prompt(backward_info: Dict[str, str]) -> str:
        conversation = CONVERSATION_TEMPLATE.format(**backward_info)
        backward_prompt = CONVERSATION_START_INSTRUCTION_CHAIN.format(conversation=conversation, **backward_info)
        backward_prompt += OBJECTIVE_INSTRUCTION_CHAIN.format(**backward_info)
        if len(backward_info['in_context_examples']) > 0:
            backward_prompt += IN_CONTEXT_EXAMPLE_PROMPT_ADDITION.format(**backward_info)
        backward_prompt += EVALUATE_VARIABLE_INSTRUCTION.format(**backward_info)
       
        return backward_prompt
    @staticmethod
    def _backward_through_llm_chain(variables: List[Variable], 
                                    response: Variable, 
                                    prompt: str, 
                                    system_prompt: str,
                                    backward_engine: EngineLM,
                                    in_context_examples: List[str]=None):
        """
        Backward through the LLM to compute gradients for each variable, in the case where the output has gradients on them.
        i.e. applying the chain rule.
        
        :param variables: The list of variables to compute gradients for.
        :type variables: List[Variable]
        :param response: The response variable.
        :type response: Variable
        :param prompt: The prompt string.
        :type prompt: str
        :param system_prompt: The system prompt string.
        :type system_prompt: str
        :param backward_engine: The backward engine to use for computing gradients.
        :type backward_engine: EngineLM

        :return: None
        """
        for variable in variables:
            if not variable.requires_grad:
                continue

            backward_info = {
                "response_desc": response.get_role_description(),
                "response_value": response.get_value(),
                "response_gradient": response.get_gradient_text(),
                "prompt": prompt,
                "system_prompt": system_prompt,
                "variable_desc": variable.get_role_description(),
                "variable_short": variable.get_short_value(),
                "in_context_examples": "\n".join(in_context_examples) if in_context_examples is not None else [],
            }
            
            backward_prompt = LLMCall_with_in_context_examples._construct_llm_chain_backward_prompt(backward_info)

            logger.info(f"_backward_through_llm prompt", extra={"_backward_through_llm": backward_prompt})
            gradient_value = backward_engine(backward_prompt, system_prompt=BACKWARD_SYSTEM_PROMPT)
            logger.info(f"_backward_through_llm gradient", extra={"_backward_through_llm": gradient_value})
            
            var_gradients = Variable(value=gradient_value, role_description=f"feedback to {variable.get_role_description()}")
            variable.gradients.add(var_gradients)
            conversation = CONVERSATION_TEMPLATE.format(**backward_info)
            variable.gradients_context[var_gradients] = {
                "context": conversation, 
                "response_desc": response.get_role_description(),
                "variable_desc": variable.get_role_description()
            }
            
            if response._reduce_meta:
                var_gradients._reduce_meta.extend(response._reduce_meta)
                variable._reduce_meta.extend(response._reduce_meta)

    @staticmethod
    def _construct_llm_base_backward_prompt(backward_info: Dict[str, str]) -> str:
        conversation = CONVERSATION_TEMPLATE.format(**backward_info)
        backward_prompt = CONVERSATION_START_INSTRUCTION_BASE.format(conversation=conversation, **backward_info)
        backward_prompt += OBJECTIVE_INSTRUCTION_BASE.format(**backward_info)
        
        if len(backward_info['in_context_examples']) > 0:
            backward_prompt += IN_CONTEXT_EXAMPLE_PROMPT_ADDITION.format(**backward_info)
        backward_prompt += EVALUATE_VARIABLE_INSTRUCTION.format(**backward_info)

        
        return backward_prompt

    @staticmethod
    def _backward_through_llm_base(variables: List[Variable], 
                                   response: Variable,
                                   prompt: str,
                                   system_prompt: str,
                                   backward_engine: EngineLM,
                                   in_context_examples: List[str]=None):
        """
        Backward pass through the LLM base. 
        In this case we do not have gradients on the output variable.

        :param variables: A list of variables to compute gradients for.
        :type variables: List[Variable]
        :param response: The response variable.
        :type response: Variable
        :param prompt: The prompt string.
        :type prompt: str
        :param system_prompt: The system prompt string.
        :type system_prompt: str
        :param backward_engine: The backward engine to use for computing gradients.
        :type backward_engine: EngineLM
        """

        for variable in variables:
            if not variable.requires_grad:
                continue

            backward_info = {
                "response_desc": response.get_role_description(),
                "response_value": response.get_value(),
                "prompt": prompt,
                "system_prompt": system_prompt,
                "variable_desc": variable.get_role_description(),
                "variable_short": variable.get_short_value(),
                "in_context_examples": "\n".join(in_context_examples) if in_context_examples is not None else [],

            }
            
            backward_prompt = LLMCall_with_in_context_examples._construct_llm_base_backward_prompt(backward_info)
            
            logger.info(f"_backward_through_llm prompt", extra={"_backward_through_llm": backward_prompt})
            gradient_value = backward_engine(backward_prompt, system_prompt=BACKWARD_SYSTEM_PROMPT)
            logger.info(f"_backward_through_llm gradient", extra={"_backward_through_llm": gradient_value})

            conversation = CONVERSATION_TEMPLATE.format(**backward_info)
            var_gradients = Variable(value=gradient_value, role_description=f"feedback to {variable.get_role_description()}")
            variable.gradients.add(var_gradients)
            variable.gradients_context[var_gradients] = {
                "context": conversation, 
                "response_desc": response.get_role_description(),
                "variable_desc": variable.get_role_description()
            }

            if response._reduce_meta:
                var_gradients._reduce_meta.extend(response._reduce_meta)
                variable._reduce_meta.extend(response._reduce_meta)


import guidance
import os

home_dir = os.path.expanduser("~")
with open(os.path.join(home_dir, "openai_api_key.txt")) as fin:
    line = fin.readline()
api_key = line.strip()

gpt = guidance.llms.OpenAI("gpt-4o-mini", api_key=api_key, chat_mode=True)

experts = guidance('''
{{#system~}}
You are a helpful and terse assistant.
{{~/system}}

{{#user~}}
I want a response to the following question:
{{query}}
Name 3 world-class experts (past or present) who would be great at answering this?
Don't answer the question yet.
{{~/user}}

{{#assistant~}}
{{gen 'expert_names' temperature=0 max_tokens=300}}
{{~/assistant}}

{{#user~}}
Great, now please answer the question as if these experts had collaborated in writing a joint anonymous answer.
{{~/user}}

{{#assistant~}}
{{gen 'answer' temperature=0 max_tokens=500}}
{{~/assistant}}
''', llm=gpt)

executed_program = experts(query='How can I be more productive?')
print(executed_program)
print(executed_program["expert_names"])
print(executed_program["answer"])


# Example output:

## executed_program["expert_names"]
# Sure! Here are three world-class experts who could provide valuable insights on productivity:                                                     

# 1. **David Allen** - Author of "Getting Things Done," a widely recognized productivity methodology.                                               
# 2. **Cal Newport** - Author of "Deep Work," focusing on the importance of focused work in achieving high productivity.                            
# 3. **Tim Ferriss** - Author of "The 4-Hour Workweek," known for his unconventional approaches to productivity and time management. 

## executed_program["answer"]
# To enhance your productivity, consider the following strategies:

# 1. **Implement a Systematic Approach**: Adopt a structured method like Getting Things Done (GTD). Break tasks into actionable steps, prioritize them, and regularly review your progress. This will help you manage your workload effectively and reduce overwhelm.                                  

# 2. **Cultivate Deep Work**: Dedicate uninterrupted blocks of time to focus on your most important tasks. Minimize distractions by creating a conducive work environment and setting clear boundaries. This will allow you to produce higher quality work in less time.                               

# 3. **Optimize Your Time and Energy**: Experiment with time management techniques, such as the Pomodoro Technique or time blocking, to find what works best for you. Additionally, assess your energy levels throughout the day and schedule demanding tasks during your peak performance times.      

# By integrating these principles, you can create a personalized productivity system that maximizes your efficiency and effectiveness. 


def is_jpeg(data):
    jpeg_signature = b'\xFF\xD8\xFF'
    return data.startswith(jpeg_signature)

def is_png(data):
    png_signature = b'\x89\x50\x4E\x47\x0D\x0A\x1A\x0A'
    return data.startswith(png_signature)


def get_image_type_from_bytes(data):
    if is_jpeg(data):
        return "jpeg"
    elif is_png(data):
        return "png"
    else:
        raise ValueError("Image type not supported, only jpeg and png supported.")


from textgrad.engine import EngineLM


def load_multimodal_instance_task(task_name: str, evaluation_api: EngineLM, *args, **kwargs):
    if task_name == "mathvista":
        from textgrad.tasks.multimodal.mathvista import MathVistaDataset
        test_set = MathVistaDataset(evaluation_api=evaluation_api, split="testmini", *args, **kwargs)
        return test_set
    
    elif task_name == "scienceqa":
        from textgrad.tasks.multimodal.scienceqa import ScienceQADataset
        test_set = ScienceQADataset(evaluation_api=evaluation_api, split="test", *args, **kwargs)
        return test_set
        
    else:
        raise ValueError(f"Instance task {task_name} not found.")



try:
    from together import Together
except ImportError:
    raise ImportError("If you'd like to use OpenAI models, please install the openai package by running `pip install together`, and add 'TOGETHER_API_KEY' to your environment variables.")

import os
import platformdirs
from tenacity import (
    retry,
    stop_after_attempt,
    wait_random_exponential,
)

from .base import EngineLM, CachedEngine

class ChatTogether(EngineLM, CachedEngine):
    DEFAULT_SYSTEM_PROMPT = "You are a helpful, creative, and smart assistant."

    def __init__(
        self,
        model_string="meta-llama/Llama-3-70b-chat-hf",
        system_prompt=DEFAULT_SYSTEM_PROMPT):
        """
        :param model_string:
        :param system_prompt:
        """
        root = platformdirs.user_cache_dir("textgrad")
        cache_path = os.path.join(root, f"cache_together_{model_string}.db")
        super().__init__(cache_path=cache_path)

        self.system_prompt = system_prompt
        if os.getenv("TOGETHER_API_KEY") is None:
            raise ValueError("Please set the TOGETHER_API_KEY environment variable if you'd like to use OpenAI models.")
        
        self.client = Together(
            api_key=os.getenv("TOGETHER_API_KEY"),
        )
        self.model_string = model_string

    def generate(
        self, prompt, system_prompt=None, temperature=0, max_tokens=2000, top_p=0.99
    ):

        sys_prompt_arg = system_prompt if system_prompt else self.system_prompt

        cache_or_none = self._check_cache(sys_prompt_arg + prompt)
        if cache_or_none is not None:
            return cache_or_none

        response = self.client.chat.completions.create(
            model=self.model_string,
            messages=[
                {"role": "system", "content": sys_prompt_arg},
                {"role": "user", "content": prompt},
            ],
            frequency_penalty=0,
            presence_penalty=0,
            stop=None,
            temperature=temperature,
            max_tokens=max_tokens,
            top_p=top_p,
        )

        response = response.choices[0].message.content
        self._save_cache(sys_prompt_arg + prompt, response)
        return response

    @retry(wait=wait_random_exponential(min=1, max=5), stop=stop_after_attempt(5))
    def __call__(self, prompt, **kwargs):
        return self.generate(prompt, **kwargs)




from .. import logger
from ..defaults import (SYSTEM_PROMPT_DEFAULT_ROLE, 
                               VARIABLE_OUTPUT_DEFAULT_ROLE)
from ..variable import Variable
from ..engine import EngineLM, validate_multimodal_engine
from typing import List, Dict
from .llm_backward_prompts import (
    EVALUATE_VARIABLE_INSTRUCTION,
    CONVERSATION_START_INSTRUCTION_BASE,
    CONVERSATION_START_INSTRUCTION_CHAIN,
    OBJECTIVE_INSTRUCTION_CHAIN,
    OBJECTIVE_INSTRUCTION_BASE,
    BACKWARD_SYSTEM_PROMPT,
)
from .multimodal_backward_prompts import MULTIMODAL_CONVERSATION_TEMPLATE
from typing import Union
from ..config import validate_engine_or_get_default
from .function import Function, BackwardContext


class MultimodalLLMCall(Function):
    """The MultiModalLM call function. This function will call the LLM with the input (image) and return the response,
    also register the grad_fn for backpropagation.

    :param engine: engine to use for the LLM call
    :type engine: EngineLM
    :param system_prompt: system prompt to use for the LLM call, default depends on the engine.
    :type system_prompt: Variable, optional
    """
    def __init__(self, 
                 engine: Union[str, EngineLM], 
                 system_prompt: Variable = None):
        super().__init__()
        self.engine = validate_engine_or_get_default(engine)
        validate_multimodal_engine(self.engine)

        self.system_prompt = system_prompt
        if self.system_prompt and self.system_prompt.get_role_description() is None:
            self.system_prompt.set_role_description(SYSTEM_PROMPT_DEFAULT_ROLE)
    
    
    def forward(self, 
                inputs: List[Variable], 
                response_role_description: str = VARIABLE_OUTPUT_DEFAULT_ROLE) -> Variable:
        """
        Forward pass for the multimodal LLM call function.

        :param inputs: list of input variables to the multimodal LLM call. One is an image and the second one is text
        :type inputs: List[Variable]
        :param response_role_description: role description for the response variable
        :type response_role_description: str, optional

        >>> from textgrad import Variable, get_engine
        >>> from textgrad.autograd import MultimodalLLMCall
        >>> target_image = "A byte representation of the image"
        >>> question_variable = Variable("What do you see here?", role_description="question to answer", requires_grad=False)
        >>> response = MultimodalLLMCall("gpt-4o")([target_image, question_variable])
        """
        # First ensure that all keys are present in the fields

        # Assert that all variables are either strings or bytes
        for variable in inputs:
            if not isinstance(variable.get_value(), (str, bytes)):
                raise ValueError(f"MultimodalLLMCall only accepts str or bytes, got {type(variable.get_value())}")
        
        system_prompt_value = self.system_prompt.value if self.system_prompt else None
        input_content = [inp.value for inp in inputs]
        # Make the LLM Call
        response_text = self.engine(input_content, system_prompt=system_prompt_value)

        # Create the response variable
        response = Variable(
            value=response_text,
            predecessors=[self.system_prompt, *inputs] if self.system_prompt else [*inputs],
            role_description=response_role_description
        )
        
        logger.info(f"MultimodalLLMCall function forward", extra={"text": f"System:{system_prompt_value}\n{inputs}"})
        
        # Populate the gradient function, using a container to store the backward function and the context
        response.set_grad_fn(BackwardContext(backward_fn=self.backward, 
                                             response=response, 
                                             input_content=input_content, 
                                             system_prompt=system_prompt_value))
        
        return response

    
    def backward(self, response: Variable, input_content: List[Union[str, bytes]], system_prompt: str, backward_engine: EngineLM):
        validate_multimodal_engine(backward_engine)

        children_variables = response.predecessors
        if response.get_gradient_text() == "":
            self._backward_through_multimodal_llm_base(children_variables, response, input_content, system_prompt, backward_engine)
        else:
            self._backward_through_multimodal_llm_chain(children_variables, response, input_content, system_prompt, backward_engine)

    @staticmethod
    def _construct_multimodal_llm_chain_backward_content(backward_info: Dict[str, str]) -> str:
        content = [c for c in backward_info["input_content"]]
        conversation = MULTIMODAL_CONVERSATION_TEMPLATE.format(**backward_info)
        backward_prompt = CONVERSATION_START_INSTRUCTION_CHAIN.format(conversation=conversation, **backward_info)
        backward_prompt += OBJECTIVE_INSTRUCTION_CHAIN.format(**backward_info)
        backward_prompt += EVALUATE_VARIABLE_INSTRUCTION.format(**backward_info)
        content.append(backward_prompt)
        return content

    @staticmethod
    def _backward_through_multimodal_llm_chain(variables: List[Variable],
                                               response: Variable,
                                               input_content: List[Union[str, bytes]],
                                               system_prompt: str,
                                               backward_engine: EngineLM):
        for variable in variables:
            if not variable.requires_grad:
                continue

            backward_info = {
                "response_desc": response.get_role_description(),
                "response_value": response.get_value(),
                "response_gradient": response.get_gradient_text(),
                "input_content": input_content,
                "system_prompt": system_prompt,
                "variable_desc": variable.get_role_description(),
                "variable_short": variable.get_short_value()
            }
            
            backward_content = MultimodalLLMCall._construct_multimodal_llm_chain_backward_content(backward_info)

            logger.info(f"_backward_through_llm prompt", extra={"_backward_through_llm": backward_content})
            gradient_value = backward_engine(backward_content, system_prompt=BACKWARD_SYSTEM_PROMPT)
            logger.info(f"_backward_through_llm gradient", extra={"_backward_through_llm": gradient_value})
            
            var_gradients = Variable(value=gradient_value, role_description=f"feedback to {variable.get_role_description()}")
            variable.gradients.add(var_gradients)
            conversation = MULTIMODAL_CONVERSATION_TEMPLATE.format(**backward_info)
            variable.gradients_context[var_gradients] = {
                "context": input_content + [conversation], 
                "response_desc": response.get_role_description(),
                "variable_desc": variable.get_role_description()
            }
            
            if response._reduce_meta:
                var_gradients._reduce_meta.extend(response._reduce_meta)
                variable._reduce_meta.extend(response._reduce_meta)

    @staticmethod
    def _construct_multimodal_llm_base_backward_content(backward_info: Dict[str, str]) -> str:
        content = [c for c in backward_info["input_content"]]
        conversation = MULTIMODAL_CONVERSATION_TEMPLATE.format(**backward_info)
        backward_prompt = CONVERSATION_START_INSTRUCTION_BASE.format(conversation=conversation, **backward_info)
        backward_prompt += OBJECTIVE_INSTRUCTION_BASE.format(**backward_info)
        backward_prompt += EVALUATE_VARIABLE_INSTRUCTION.format(**backward_info)
        content.append(backward_prompt)
        return content

    @staticmethod
    def _backward_through_multimodal_llm_base(variables: List[Variable],
                                              response: Variable,
                                              input_content: List[Union[str, bytes]],
                                              system_prompt: str,
                                              backward_engine: EngineLM):
        for variable in variables:
            if not variable.requires_grad:
                continue

            backward_info = {
                "response_desc": response.get_role_description(),
                "response_value": response.get_value(),
                "input_content": input_content,
                "system_prompt": system_prompt,
                "variable_desc": variable.get_role_description(),
                "variable_short": variable.get_short_value()
            }
            
            backward_content = MultimodalLLMCall._construct_multimodal_llm_base_backward_content(backward_info)
            
            logger.info(f"_backward_through_llm prompt", extra={"_backward_through_llm": backward_content})
            gradient_value = backward_engine(backward_content, system_prompt=BACKWARD_SYSTEM_PROMPT)
            logger.info(f"_backward_through_llm gradient", extra={"_backward_through_llm": gradient_value})

            conversation = MULTIMODAL_CONVERSATION_TEMPLATE.format(**backward_info)
            var_gradients = Variable(value=gradient_value, role_description=f"feedback to {variable.get_role_description()}")
            variable.gradients.add(var_gradients)
            variable.gradients_context[var_gradients] = {
                "context": input_content + [conversation], 
                "response_desc": response.get_role_description(),
                "variable_desc": variable.get_role_description()
            }

            if response._reduce_meta:
                var_gradients._reduce_meta.extend(response._reduce_meta)
                variable._reduce_meta.extend(response._reduce_meta)



class OrderedFieldsMultimodalLLMCall(MultimodalLLMCall):
    def __init__(self, 
                 engine: Union[str, EngineLM], 
                 fields: List[str],
                 system_prompt: Variable = None):

        self.engine = validate_engine_or_get_default(engine)
        validate_multimodal_engine(self.engine)

        self.system_prompt = system_prompt
        if self.system_prompt and self.system_prompt.get_role_description() is None:
            self.system_prompt.set_role_description(SYSTEM_PROMPT_DEFAULT_ROLE)

        self.fields = fields
        
    def forward(self, 
                inputs: Dict[str, Variable],
                response_role_description: str = VARIABLE_OUTPUT_DEFAULT_ROLE) -> Variable:
        # Assert that all variables are either strings or bytes
        for variable in inputs.values():
            if not isinstance(variable.get_value(), (str, bytes)):
                raise ValueError(f"MultimodalLLMCall only accepts str or bytes, got {type(variable.get_value())}")
        
        assert set(inputs.keys()) == set(self.fields), f"Expected fields {self.fields.keys()} but got {inputs.keys()}"
        forward_content = []
        for field in self.fields:
            if type(inputs[field].value) == bytes:
                forward_content.append(inputs[field].value)
            else:
                forward_content.append(f"{field}: {inputs[field].value}")
        
        system_prompt_value = self.system_prompt.value if self.system_prompt else None

        # Make the LLM Call
        response_text = self.engine(forward_content, system_prompt=system_prompt_value)

        # Create the response variable
        response = Variable(
            value=response_text,
            predecessors=[self.system_prompt, *(list(inputs.values()))] if self.system_prompt else [*(list(inputs.values()))],
            role_description=response_role_description
        )
        
        logger.info(f"MultimodalLLMCall function forward", extra={"text": f"System:{system_prompt_value}\n{forward_content}"})
        
        # Populate the gradient function, using a container to store the backward function and the context
        response.set_grad_fn(BackwardContext(backward_fn=self.backward, 
                                             response=response, 
                                             input_content=forward_content, 
                                             system_prompt=system_prompt_value))
        
        return response



from typing import Tuple

from transformers import AutoModelForCausalLM, Gemma2ForCausalLM, LlamaForCausalLM


def model_supported(model: AutoModelForCausalLM) -> Tuple[bool, str]:
    if isinstance(model, LlamaForCausalLM):
        return True, "Llama3"
    elif isinstance(model, Gemma2ForCausalLM):
        return True, "Gemma2"
    else:
        return False, None


def llama_post_process(text: str) -> str:
    text = text.split("<|end_header_id|>")[-1]
    text = text.replace("<|eot_id|>","")
    text = text.replace("<|begin_of_text|>", "")
    
    return text



import numpy as np
from abc import ABC, abstractmethod

class Dataset(ABC):
    @abstractmethod
    def __init__(self):
        pass
    
    @abstractmethod
    def __getitem__(self):
        pass
    
    @abstractmethod
    def __len__(self):
        pass
    
    @abstractmethod
    def get_default_task_instruction(self):
        pass

class DataLoader:
    def __init__(self, data, batch_size=32, shuffle=True):
        self.data = data
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.indices = np.arange(len(data))
        self.current_index = 0

    def __iter__(self):
        if self.shuffle:
            np.random.shuffle(self.indices)
        self.current_index = 0
        return self

    def __next__(self):
        if self.current_index >= len(self.data):
            raise StopIteration

        batch_indices = self.indices[self.current_index : self.current_index + self.batch_size]
        batch_data = [self.data[int(i)] for i in batch_indices]

        if isinstance(batch_data[0], tuple):
            batch_data = tuple(zip(*batch_data))
        else:
            batch_data = np.array(batch_data)

        self.current_index += self.batch_size

        return batch_data


# This is how we save gradients to the variable.
GRADIENT_TEMPLATE = (
    "Here is a conversation:\n\n<CONVERSATION>{context}</CONVERSATION>\n\n"
    "This conversation is potentially part of a larger system. The output is used as {response_desc}\n\n"
    "Here is the feedback we got for {variable_desc} in the conversation:\n\n<FEEDBACK>{feedback}</FEEDBACK>\n\n"
)



import os
import tqdm
import retrying

import openai

from .model import AbstractModel
from .utils import _load_prompt, create_batches


FULL_PRMOPT = """
Instruction: {{instruction}}

Input: {{input}}
Output: 
"""

class DirectModel(AbstractModel):
    def load_prompt(self):
        self.full_prompt = FULL_PRMOPT
  
    @retrying.retry(
        stop_max_attempt_number=100, # Maximum number of attempts
        wait_exponential_multiplier=1000, wait_exponential_max=10000, 
        # Wait 2^x * 1000 milliseconds between each retry, up to 10 seconds, then 10 seconds afterwards
        retry_on_exception=lambda e: isinstance(e, Exception)
    )
    def _request_openai(self, queries):
        engine = "text-davinci-003" if self.args.task_model == "openai_td003" else "gpt-3.5-turbo-instruct"

        responses = openai.Completion.create(
            engine=engine,
            prompt=queries,
            max_tokens=400,
            temperature=0.0,
            n=1,
            top_p=1,
            stream=False,
            stop=None
        )
        
        assert len(responses.choices) == len(queries)
        
        all_raw_output = [choice.text.strip() for choice in responses.choices]
        
        return all_raw_output
        
    
    def _run_batch(self, instruction, batch):
        all_inputs = batch["input"].tolist()
        all_queries = []
        
        prompt = self.full_prompt.replace("{{instruction}}", instruction)
        
        for i in range(len(all_inputs)):
            query = prompt.replace("{{input}}", all_inputs[i])
            all_queries.append(query)
                               
        # response = self._request_substrate(all_queries)
        # all_raw_output = [choice["text"] for choice in response["choices"]]
        all_raw_output = self._request_openai(all_queries)
        
        return all_raw_output
    
    def run(self, instruction, data):
        data = data.copy()
        batches = create_batches(data, batch_size=10)
        
        all_raw_output = []
        for batch in tqdm.tqdm(batches, desc="Batches"):
            batch_raw_output = self._run_batch(instruction, batch)
            all_raw_output += batch_raw_output

        data["raw_output"] = all_raw_output

        return data
    
    def run_till_enough_errors(self, instruction, data, task, target_n_errors):
        data = data.copy().sample(frac=1.0, random_state=42) # random shuffle
        batches = create_batches(data, batch_size=10)
        
        all_raw_output = []
        n_done = 0

        for batch in tqdm.tqdm(batches, desc="Batches"):
            batch_raw_output = self._run_batch(instruction, batch)
            
            n_done += len(batch)
            all_raw_output += batch_raw_output

            data_done = data.iloc[:n_done]
            data_done["raw_output"] = all_raw_output

            data_done, _ = task.evaluate(data_done)
            n_errors = (data_done['score'] == 0.0).sum()
            if n_errors >= target_n_errors:
                self.logger.info("Found {} model mistakes by running {} examples".format(n_errors, n_done))
                break

        return data_done



from .direct import DirectTask
from .zeroshotcot import ZeroshotCoTTask

TASK2CLASS = {
    "ii": DirectTask,
    "math": ZeroshotCoTTask,
    "bbh": ZeroshotCoTTask,
    "cf": DirectTask,
    "custom": ZeroshotCoTTask,
}

def Task2Class(task):
    if task not in TASK2CLASS:
        raise NotImplementedError
    else:
        return TASK2CLASS[task]



from src.greaterprompt import (
    ApeOptimizer, ApoOptimizer, GreaterOptimizer, GreaterDataloader, Pe2Optimizer, TextGradOptimizer
)

import torch
from torch.nn import functional as F
from transformers import AutoModelForCausalLM, AutoTokenizer


# ------ Part 1: Build your dataloader ------ #

# method1: use jsonl file to build the dataloader
dataloader1 = GreaterDataloader(data_path="./data/boolean_expressions.jsonl")

# method2: use custom inputs to build the dataloader
dataloader2 = GreaterDataloader(custom_inputs=[
    {
        "question": "((-1 + 2 + 9 * 5) - (-2 + -4 + -4 * -7)) =", 
        "prompt": "Use logical reasoning and think step by step.", 
        "answer": "24"
    },
    {
        "question": "((-9 * -5 - 6 + -2) - (-8 - -6 * -3 * 1)) =",
        "prompt": "Use logical reasoning and think step by step.",
        "answer": "63"
     },
    {
        "question": "((3 * -3 * 6 + -5) - (-2 + -7 - 7 - -7)) =",
        "prompt": "Use logical reasoning and think step by step.",
        "answer": "-50"
    }
])

# ------ Part 2: Initialize the model and tokenizer for GreaterOptimizer only------ #

# init model and tokenzier if you want to use GreaterOptimizer
MODEL_PATH = "/scratch1/wmz5132/models/huggingface/gemma-2-9b-it"
DEVICE = "cuda"
model = AutoModelForCausalLM.from_pretrained(MODEL_PATH, torch_dtype=torch.bfloat16, device_map=DEVICE)
model.gradient_checkpointing_enable()
tokenizer = AutoTokenizer.from_pretrained(MODEL_PATH)

# ------ Part 3: Set the optimize configs (Optional) ------ # 

# for GreaterOptimizer
greater_optimize_config = {
    "intersect_q": 5,
    "candidates_topk": 10,
    "loss_function": F.cross_entropy,
    "perplexity_loss": True,
    "perplexity_lambda": 0.2,
    "filter": True,
    "generate_config": {
        "do_sample": True,
        "temperature": 0.2,
        "max_new_tokens": 512,
    }
}

# for ApeOptimizer, ApoOptimizer, Pe2Optimizer
optimize_config = {
    "task_model": "openai_gpt35_turbo_instruct",
    "optim_model": "openai_gpt4_turbo",
}

# for TextGradOptimizer
textgrad_optimize_config = {
    "evaluation_engine": "/scratch1/wmz5132/models/huggingface/Llama-3.1-8B-Instruct",
    "test_engine": "/scratch1/wmz5132/models/huggingface/Llama-3.1-8B-Instruct",
    "device": "cuda"
}

# ------ Part 4: Initialize the Optimizers ------ #

ape_optimizer = ApeOptimizer(
    optimize_config=optimize_config # Optional
)

apo_optimizer = ApoOptimizer(
    optimize_config=optimize_config # Optional
)

greater_optimizer = GreaterOptimizer(
    model=model,
    tokenizer=tokenizer,
    optimize_config=greater_optimize_config # Optional
)

pe2_optimizer = Pe2Optimizer(
    optimize_config=optimize_config # Optional
)

textgrad_optimizer = TextGradOptimizer(
    textgrad_optimize_config # Optional
)

# ------ Part 5: Pass the Dataloader to Optimizer and Optimize ------ #

ape_result = ape_optimizer.optimize(dataloader1, p_init="think step by step")
apo_result = apo_optimizer.optimize(dataloader2, p_init="think step by step")
greater_result = greater_optimizer.optimize(
    dataloader1,
    p_extractor="\nNext, only give the exact answer, no extract words or any punctuation:",
    rounds=1
)
pe2_result = pe2_optimizer.optimize(dataloader2, p_init="think step by step")
textgrad_result = textgrad_optimizer.optimize(dataloader1, p_init="think step by step")

# print results
print(f'ape_result: {ape_result}')
print(f'-' * 30)
print(f'apo_result: {apo_result}')
print(f'-' * 30)
print(f'greater_result: {greater_result}')
print(f'-' * 30)
print(f'pe2_result: {pe2_result}')
print(f'-' * 30)
print(f'textgrad_result: {textgrad_result}')



import json
from typing import Any, Dict, List

from torch.utils.data import Dataset


class GreaterDataloader(Dataset):
    def __init__(
        self, custom_inputs: List[dict] | None = None, data_path: str | None = None, **kwargs
    ):
        super().__init__()

        self.custom_inputs = custom_inputs
        self.data_path = data_path
        self.items = []

        self._load_data() if data_path else self._build_data()


    def _load_data(self):
        with open(self.data_path, "r") as f:
            self.items = [json.loads(line) for line in f]


    def _build_data(self):
        for i, line in enumerate(self.custom_inputs):
            self.items.append({
                "id": line.get("id", str(i)),
                "question": line["question"],
                "prompt": line["prompt"],
                "answer": line["answer"]
            })


    def __len__(self) -> int:
        return len(self.items)
    

    def __getitem__(self, idx: int) -> Dict[str, Any]:
        return self.items[idx]



import os
import argparse
import logging

from .run import run
from .utils import get_logger, seed_everything

def ape_apo_pe2_optimizer(args):
    # create output_dir
    if not os.path.exists(args.output_dir):
        os.makedirs(args.output_dir, exist_ok=True)

    # create logger
    logger = get_logger(args)
    logger.info(args)
    
    # seed everything
    seed_everything(args.seed)

    p_start = run(args, logger)

    return p_start



import os
import pickle
import guidance
import glob
import random
import itertools
import openai

from .trainer import AbstractTrainer
from .utils import deduplicate, clean_string, get_llm, pack_demo_string, _load_prompt
from .node import Node

class DefaultTrainer(AbstractTrainer):
    
    def __init__(self, args, logger):
        super().__init__(args, logger)
        self.llm = get_llm(self.args.optim_model)
        self.prepare_meta_prompts()

    def initialize(self, task):
        train_data = task.get_data_split("train") # training data
        
        # get a list of initial prompts
        if self.args.init_method == "induction":
            prompts = self.initialize_induction(train_data)
            self.logger.info("Initialization prompts: {}".format(prompts))
        elif self.args.init_method == "file":
            prompts = self.initialize_file()
            
        # pack them into nodes for optimization
        initial_nodes = self._pack_initial_nodes(prompts)
        return initial_nodes
    
    def initialize_file(self):
        prompt_files = glob.glob(os.path.join(self.args.data_dir, self.args.init_prompt_file))
        all_prompts = []
        for filename in prompt_files:
            prompt = _load_prompt(filename)
            all_prompts.append(prompt)
        return all_prompts
    
    def initialize_induction(self, train_data):

        init_batch, _ = self._sample_batch(train_data, k=self.args.init_n_demo, method="random")
        demo_string = pack_demo_string(init_batch)
        
        f = self.initializer(demos=demo_string, n_demo=self.args.init_n_demo, max_tokens=self.args.prompt_max_tokens)
        init_instructions = f["instruction"]
        
        # when n=1 guidance returns a string; otherwise it returns a list of strings
        if self.args.init_n_prompts == 1:
            return [clean_string(init_instructions)]
        else:
            return [clean_string(item) for item in init_instructions]
        
    def train(self, model, initial_nodes, task):
        self.model = model
        
        train_data = task.get_data_split("train") # training data
        dev_data = task.get_data_split("dev") # dev data
        
        T = self.args.train_steps # total optimization steps
        
        # initialization/loading states
        checkpoint_path = os.path.join(self.args.output_dir, "checkpoint.pkl")
        if os.path.exists(checkpoint_path) and self.args.resume:
            states, t = self.load_session(checkpoint_path)
            self.logger.info("Resuming from checkpoint {}; t={}".format(checkpoint_path, t))
        else:
            states, t = [initial_nodes], 0
            self.logger.info("Start training from scratch".format(checkpoint_path))
            
        stop_training = False

        # training
        while t < T:
            
            # evaluate all prompt candidates
            for i, node in enumerate(states[-1]):
                
                if self.args.do_validate:
                    dev_results, dev_score = self.evaluate(model, node, task, dev_data, "dev")
                    self.logger.info("[Node {}] Dev Score: {}".format(node.id, dev_score))
                    
                node.save(self.args.output_dir)
                
                if self.args.do_validate and dev_score == 1.0:
                    stop_training = True
                    break
                
                # save checkpoint (save evaluation progress)
                self.save_session(checkpoint_path, states, t)
                
            if stop_training:
                break
                
            # select top k prompts
            selection_criterion = "dev"
            # `backtracking`: selects best states globally
            candidates = list(itertools.chain(*states)) if self.args.backtrack else states[-1]
            selected_nodes = sorted(candidates, key=lambda x: x.scores.get(selection_criterion, -1.0), reverse=True)[:self.args.n_beam] # select best
            self.logger.info("[t={}] Selected nodes: {}".format(t, str([node.id for node in selected_nodes])))
            
            # at the last timestamp do not expand new prompt candidates
            if t == T-1:
                break
            
            # expand new prompt candidates
            new_nodes = []
            for node in selected_nodes:
                _, _ = self.evaluate(model, node, task, train_data, "train")
                new_nodes += self._propose_new_nodes(node)
                    
            # deduplicate
            new_nodes = deduplicate(new_nodes, prev_nodes=list(itertools.chain(*states)))
            states.append(new_nodes)

            # increment the timestamp
            t += 1
            
            # save checkpoint
            self.save_session(checkpoint_path, states, t)
        
        # find the best node
        selection_criterion = "dev" if self.args.do_validate else "train"
        candidates = list(itertools.chain(*states))
        best_node = sorted(candidates, key=lambda x: x.scores.get(selection_criterion, -1.0), reverse=True)[0] # select best
        best_score = best_node.scores[selection_criterion]  

        return best_score, best_node, states
    
    def evaluate(self, model, node, task, data, split_name):
        # data is a dataframe
        
        # the node may be evaluated before the code crashes;
        # in this case we don't need to run evaluation again.
        if split_name not in node.scores:

            if split_name == "train":
                n_target_errors = self.args.batch_size * max(self.args.n_expand, 2) / 2 # ensuring enough error cases for generating feedback
                data = model.run_till_enough_errors(node.prompt, data, task, n_target_errors)
            else:
                data = model.run(node.prompt, data)
                
            data, final_score = task.evaluate(data)
            node.register_results(data, final_score, split_name)
            if self.args.optim_use_momentum and split_name == "train":
                node.history += "\n * At timestamp {}, the accuracy on the training set is {:.4f}".format(
                    node.timestamp, final_score
                )

        return node.results[split_name], node.scores[split_name]
    
    def load_session(self, checkpoint_file):
        with open(checkpoint_file, "rb") as f:
            states, t = pickle.load(f)
        return states, t
            
    def save_session(self, checkpoint_file, states, t):
        with open(checkpoint_file, "wb") as f:
            pickle.dump((states, t), f)

    def _propose_new_nodes(self, node):
        new_nodes = []
        for id in range(self.args.n_expand):
            # sample batch
            batch, indices = self._sample_batch(node.results["train"], k=self.args.batch_size, method=self.args.batching)
            new_node = self._update(node, batch, indices)
            new_node.save(self.args.output_dir)
            new_nodes.append(new_node)
        return new_nodes
    
    def _update(self, node, batch, indices):
        raise NotImplementedError

    def _pack_batch(self, batch):
        s = ""
        if self.args.task in ["math", "bbh"]:
            for i, row in batch.iterrows():
                s += "### Example {}\n".format(i)
                s += "Input: {}\n".format(row["input"])
                s += "Reasoning: {}\n".format(row["reasoning"])
                s += "Output: {}\n".format(row["output"])
                s += "Label: {}\n\n".format(row["label"])
        elif self.args.task == "ii":
            for i, row in batch.iterrows():
                s += "### Example {}\n".format(i)
                s += "Input: {}\n".format(row["input"])
                s += "Output: {}\n".format(row["output"])
                # II tasks may have multiple correct answers
                s += "All Correct Labels: {}\n\n".format(row["label"])
        elif self.args.task == "cf":
            for i, row in batch.iterrows():
                s += "### Example {}\n".format(i)
                s += "Input: {}\n".format(row["input"])
                s += "Output: {}\n".format(row["output"])
                s += "Label: {}\n\n".format(row["label"][0])
        return s
    
    def _pack_initial_nodes(self, prompts):
        nodes = []
        for i, prompt in enumerate(prompts):
            node = Node(
                timestamp=0,
                id=str(i), 
                prompt=prompt
            )
            nodes.append(node)
        return nodes
    
    def _sample_batch(self, results, k, method="hard"):
        if method == "random":

            indices = results.index.tolist()
            random_indices = random.sample(indices, k=k)

            selected_examples = results.loc[random_indices]
            return selected_examples, random_indices

        elif method == "hard":

            failure_indices = results[results['score'] == 0.0].index.tolist()
            true_k = min(k, len(failure_indices)) # in the rare case that k < n_failures

            random_indices = random.sample(failure_indices, k=true_k)
            selected_examples = results.loc[random_indices]
            return selected_examples, random_indices
        
        elif method == "hard_weighted":
            
            failure_indices = results[results['score'] == 0.0].index.tolist()
            failure_scores = 1.0 - results.loc[failure_indices]['score'] # lower score -> higher prob of being selected
            true_k = min(k, len(failure_indices))
        
            # seems that there is no function for weighted unique sampling...
            random_indices = random.choices(failure_indices, weights=failure_scores, k=true_k) 
            selected_examples = results.loc[random_indices]
            return selected_examples, random_indices
            
            


import guidance
import openai

def get_llm(llm_name):
    assert llm_name in ["openai_gpt35", "openai_gpt4", "openai_td003", "openai_gpt35_turbo_instruct"]
    
    if llm_name == "openai_gpt35":
        llm = guidance.llms.OpenAI("gpt-3.5-turbo", max_retries=100, max_calls_per_min=10)
    if llm_name == "openai_gpt35_turbo_instruct":
        llm = guidance.llms.OpenAI("gpt-3.5-turbo-instruct", max_retries=100, max_calls_per_min=10)
    elif llm_name == "openai_td003":
        llm = guidance.llms.OpenAI("text-davinci-003", max_retries=100, max_calls_per_min=10)
    elif llm_name == "openai_gpt4":
        llm = guidance.llms.OpenAI("gpt-4", max_retries=100, max_calls_per_min=5)

    return llm

def create_batches(df, batch_size):
    # Calculate the number of batches
    num_batches = len(df) // batch_size

    # Create batches
    batches = [df.iloc[i * batch_size : (i + 1) * batch_size] for i in range(num_batches)]

    # If there are remaining rows not included in full batches, you can add them to the last batch
    remaining_rows = len(df) % batch_size
    if remaining_rows > 0:
        batches.append(df.iloc[-remaining_rows:])
        
    return batches

def _load_prompt(datafile: str):
    with open(datafile, "r") as f:
        prompt = f.read()
    return prompt

def clean_string(s):
    # issues with substrate llm api
    return s.replace("<|im_start|>", "").replace("<|im_end|>", "").strip()

def last_boxed_only_string(string):
    # from https://github.com/hendrycks/math/blob/main/modeling/dataset/util.py
    # for MATH dataset
    idx = string.rfind("\\boxed")
    if idx < 0:
        idx = string.rfind("\\fbox")
        if idx < 0:
            return None

    i = idx
    right_brace_idx = None
    num_left_braces_open = 0
    while i < len(string):
        if string[i] == "{":
            num_left_braces_open += 1
        if string[i] == "}":
            num_left_braces_open -= 1
            if num_left_braces_open == 0:
                right_brace_idx = i
                break
        i += 1
    
    if right_brace_idx == None:
        retval = None
    else:
        retval = string[idx:right_brace_idx + 1]
    
    return retval

def get_answer_cleansing_prompt(args):
    if args.task == "math":
        prompt = "Therefore, the answer (arabic numerals) is"
    elif args.task == "bbh":
        if args.subtask == "formal_fallacies":
            prompt = "Therefore, the answer (valid or invalid) is"
        elif args.subtask in ["causal_judgement", "sports_understanding", "web_of_lies", "navigate"]:
            prompt = "Therefore, the answer (yes or no) is"
        else:
            prompt = "Therefore, the answer is"
    elif args.task == "custom":
        prompt = "Therefore, the answer is"
    else:
        raise NotImplementedError
    return prompt


from typing import List, Dict

from .. import Variable
from ..engine import EngineLM
from .algebra import Sum, Aggregate
from .llm_ops import LLMCall, FormattedLLMCall

def sum(variables: List[Variable]) -> Variable:
    """
    Represents a sum operation on a list of variables. 
    In TextGrad, sum is simply concatenation of the values of the variables.

    :param variables: The list of variables to be summed (concatenated).
    :type variables: List[Variable]
    :return: A new variable representing the sum of the input variables.
    :rtype: Variable
    """
    return Sum()(variables)


def aggregate(variables: List[Variable]) -> Variable:
    """
    WIP - Aggregates a list of variables.
    In TextGrad, forward pass of aggregation is simply concatenation of the values of the variables.
    The backward pass performs a reduction operation on the gradients of the variables.
    This reduction is currently an LLM call to summarize the gradients.

    :param variables: The list of variables to be aggregated.
    :type variables: List[Variable]
    :return: The aggregated variable.
    :rtype: Variable
    """
    return Aggregate()(variables)


def llm_call(input_variable: Variable, engine: EngineLM, 
             response_role_description: str = None, system_prompt: Variable = None):
    """A functional version of the LLMCall.
    The simple LLM call function. This function will call the LLM with the input and return the response, also register the grad_fn for backpropagation.
    
    :param input_variable: The input variable (aka prompt) to use for the LLM call.
    :type input_variable: Variable
    :param response_role_description: Role description for the LLM response, defaults to VARIABLE_OUTPUT_DEFAULT_ROLE
    :type response_role_description: str, optional
    :return: response sampled from the LLM
    :rtype: Variable
    :param engine: engine to use for the LLM call
    :type engine: EngineLM
    :param input_role_description: role description for the input variable, defaults to VARIABLE_INPUT_DEFAULT_ROLE
    :type input_role_description: str, optional
    :param system_prompt: system prompt to use for the LLM call, default depends on the engine.
    :type system_prompt: Variable, optional
    
    >>> from textgrad import Variable, get_engine
    >>> from textgrad.autograd.functional import llm_call
    >>> engine = get_engine("gpt-3.5-turbo")
    >>> prompt = Variable("What is the capital of France?", role_description="prompt to the LM")
    >>> response = llm_call(prompt, engine=engine) 
    # This returns something like Variable(data=The capital of France is Paris., grads=)
    
    """
    return LLMCall(engine=engine, system_prompt=system_prompt)(input_variable, response_role_description)


def formatted_llm_call(inputs: List[Variable], response_role_description: str, 
                       engine: EngineLM, format_string: str, 
                       fields: Dict[str, str], system_prompt: Variable = None):
    """A functional version of the LLM call with formatted strings. 
    Just a wrapper around the FormattedLLMCall class.
    
    This function will call the LLM with the input and return the response, also register the grad_fn for backpropagation.

    :param inputs: Variables to use for the input. This should be a mapping of the fields to the variables.
    :type inputs: dict[str, Variable]
    :param response_role_description: Role description for the response variable, defaults to VARIABLE_OUTPUT_DEFAULT_ROLE
    :type response_role_description: str, optional
    :param engine: The engine to use for the LLM call.
    :type engine: EngineLM
    :param format_string: The format string to use for the input. For instance, "The capital of {country} is {capital}". For a format string like this, we'll expect to have the fields dictionary to have the keys "country" and "capital". Similarly, in the forward pass, we'll expect the input variables to have the keys "country" and "capital".
    :type format_string: str
    :param fields: The fields to use for the format string. For the above example, this would be {"country": {}, "capital": {}}. This is currently a dictionary in case we'd want to inject more information later on.
    :type fields: Dict[str, str]
    :param system_prompt: The system prompt to use for the LLM call. Default value depends on the engine.
    :type system_prompt: Variable, optional
    :return: Sampled response from the LLM
    :rtype: Variable
    """
    call_object = FormattedLLMCall(engine=engine, format_string=format_string, fields=fields, system_prompt=system_prompt)
    return call_object(inputs, response_role_description)


import os
import json
import ast

import numpy as np
import pandas as pd

import random

from .task import AbstractTask
from .ii_utils import TASK_TO_METRIC, score_fn

class DirectTask(AbstractTask):
    
    def __init__(self, args, logger, data_dir):
        AbstractTask.__init__(self, args, logger, data_dir)
        self.sample_id = str(data_dir.split("/")[-1])

    def _load_split(self, split):
        filename = os.path.join(self.data_dir, "{}.csv".format(split))
        data, size = None, 0
            
        if os.path.exists(filename):
            data = pd.read_csv(filename, index_col=0)
            data.fillna('', inplace=True)
            
            indices, inputs, outputs = [], [], []
            for i, row in data.iterrows():
                index = row.name
                # breakpoint()
                # adapted from https://github.com/keirp/automatic_prompt_engineer/blob/main/experiments/data/instruction_induction/load_data.py
                if self.subtask == 'cause_and_effect':
                    cause, effect = row['cause'], row['effect']
                    # Pick an order randomly
                    if random.random() < 0.5:
                        input_ = f'Sentence 1: {cause} Sentence 2: {effect}'
                    else:
                        input_ = f'Sentence 1: {effect} Sentence 2: {cause}'
                    output_ = [cause]
                elif self.subtask == 'common_concept':
                    items = ast.literal_eval(row['items'])
                    # Make comma separated list of items
                    input_ = ', '.join(items) # ? why does the original code has items[:-1]?
                    output_ = ast.literal_eval(row['all_common_concepts'])
                elif self.subtask == 'rhymes':
                    input_, output_ = row['input'], ast.literal_eval(row['other_rhymes'])
                elif 'translation' in self.subtask:
                    input_, output_ = row['input'], ast.literal_eval(row['possible_translations'])
                elif self.subtask in ["sum", "diff"]:
                    input_, output_ = row['input'], [str(row['output'])] # do exact match for strings
                elif self.subtask == "num_to_verbal":
                    input_, output_ = str(row['input']), [row['output']] # do exact match for strings
                elif "arithmetic_base" in self.subtask:
                    input_, output_ = row['input'], [str(row['output'])] # do exact match for strings
                else:
                    input_, output_ = row['input'], [row['output']]
                
                indices.append(index)
                inputs.append(input_)
                outputs.append(output_) # output_ is a list of acceptable answers
            
            data_dict = {'input': inputs, 'label': outputs, 'index': indices}
            data = pd.DataFrame(data_dict).set_index('index')

            if self.args.debug and split in ["train", "dev"]:
                data = data.sample(n=5, random_state=self.args.seed)
                self.logger.info("Loading 5 {} examples in debug mode".format(split))
                
            size = len(data)
        
        return data, size
            
    def load_data(self):
        self.train_data, self.train_size = self._load_split("train")
        self.dev_data, self.dev_size = self._load_split("dev")
        self.test_data, self.test_size = self._load_split("test")
        
        self.data["train"] = self.train_data
        self.data["dev"] = self.dev_data
        self.data["test"] = self.test_data

        self.logger.info("Loading data... # Train: {}; # Dev: {}; # Test: {}".format(self.train_size, self.dev_size, self.test_size))
        
    def evaluate(self, result_df):
        labels = result_df["label"].tolist()
        outputs = result_df["raw_output"].tolist()
        result_df["output"] = result_df["raw_output"] # to ensure there is a column named output, will be used in packing batches
        metric = TASK_TO_METRIC.get(self.subtask, "em")

        scores = []
        for prediction, ans_ in zip(outputs, labels):
            score = score_fn(prediction, ans_, metric=metric)
            scores.append(score)
            
        result_df["score"] = scores
        return result_df, result_df["score"].mean()
    
    def __str__(self):
        s = "Task: {}; ".format(self.task)
        s += "Subtask: {}; ".format(self.subtask)
        s += "Sample ID: {}; ".format(self.sample_id)
        s += "Data Dir: {}".format(self.data_dir)
        return s




# from openai import OpenAI
import openai
import os

home_dir = os.path.expanduser("~")
with open(os.path.join(home_dir, "openai_api_key.txt")) as fin:
    line = fin.readline()
api_key = line.strip()
print(api_key)

openai.api_key=api_key

output = openai.ChatCompletion.create(
    model="gpt-4o-mini",
    messages=[{"role": "user", "content": "Say this is a test"}],
    stream=False,
)
print(output)

## Example output:
# {
#   "id": "chatcmpl-xxx",
#   "object": "chat.completion",
#   "created": 1722380151,
#   "model": "gpt-4o-mini-2024-07-18",
#   "choices": [
#     {
#       "index": 0,
#       "message": {
#         "role": "assistant",
#         "content": "This is a test."
#       },
#       "logprobs": null,
#       "finish_reason": "stop"
#     }
#   ],
#   "usage": {
#     "prompt_tokens": 12,
#     "completion_tokens": 5,
#     "total_tokens": 17
#   },
#   "system_fingerprint": "xxx"
# }


# First part of the prompt for the llm backward function
MULTIMODAL_CONVERSATION_TEMPLATE = (
    "\n Above messages are the <LM_INPUT>\n\n"
    "<LM_SYSTEM_PROMPT> {system_prompt} </LM_SYSTEM_PROMPT>\n\n"
    "<LM_OUTPUT> {response_value} </LM_OUTPUT>\n\n"
)



import os
import argparse
import logging

from run import run
from utils import get_logger, seed_everything

def main():
    parser = argparse.ArgumentParser(description="language-based optimizer")

    # basic
    parser.add_argument("--output_dir", default="output/test", type=str)
    parser.add_argument('--resume', action='store_true', help="resume training/searching from the checkpoint if there is one in output_dir")
    parser.add_argument('--do_train', action='store_true')
    parser.add_argument('--do_validate', action='store_true')
    parser.add_argument('--do_test', action='store_true')
    
    # task
    parser.add_argument("--task", default=None, required=True)
    parser.add_argument("--subtask", default=None)
    parser.add_argument("--data_dir", default=None, required=True)
    parser.add_argument("--model", choices=["direct", "zeroshotcot"], required=True)

    # llms
    parser.add_argument("--task_model", default="openai_td003")
    parser.add_argument("--optim_model", default="openai_gpt4")
    
    # initialize
    parser.add_argument("--init_method", default="file", choices=["file", "induction"])
    ## used when init method is file
    parser.add_argument("--init_prompt_file", default="prompt.md", type=str, help="the file should be in `data_dir`, supports regex matching.")
    ## used when init method is induction
    parser.add_argument("--init_n_demo", default=None, type=int, help="number of examples used in instruction induction to generate initial prompts.")
    parser.add_argument("--init_n_prompts", default=1, type=int, help="number of initial prompt candidates.")
    parser.add_argument("--init_temperature", default=0.0, type=float, help="number of initial prompt candidates.")
    parser.add_argument("--prompt_max_tokens", default=50, type=int, help="max number of tokens in a prompt")

    # optim
    parser.add_argument("--train_steps", default=10, type=int, help="# number of optimization steps")
    parser.add_argument("--batch_size", default=2, type=int, help="# failure examples included in one optimizaiton step")
    parser.add_argument("--step_size", default=10, type=int, help="# tokens that can be changed during on optimization step")
    parser.add_argument("--n_beam", default=2, type=int, help="beam size in beam search")
    parser.add_argument("--n_expand", default=2, type=int, help="# prompts to expand at each time stamp")
    
    # optim - guidance
    parser.add_argument("--meta_prompts_dir", default="meta_prompts/default", type=str)
    parser.add_argument("--optim_use_gradient", action='store_true')
    parser.add_argument("--optim_use_step_size", action='store_true')
    parser.add_argument("--optim_use_momentum", action='store_true')
    parser.add_argument("--optim_use_instruction", action='store_true')
    parser.add_argument("--optim_use_demonstrations", action='store_true')
    parser.add_argument("--optim_use_optim_tutorial", action='store_true')
    
    # search (TODO)
    parser.add_argument("--trainer", default="default", choices=["default", "ape", "apo", "pe2"])
    parser.add_argument("--batching", default="hard", choices=["random", "hard", "hard_weighted"])
    parser.add_argument("--bandit", default="all", choices=["all", "ucb"])
    parser.add_argument("--backtrack", action='store_true')
    
    # others
    parser.add_argument('--seed', type=int, default=42,
                        help="random seed for initialization")
    parser.add_argument('--debug', action='store_true',
                        help="Use a subset of data for debugging")

    # parse args
    args = parser.parse_args()

    # create output_dir
    if not os.path.exists(args.output_dir):
        os.makedirs(args.output_dir, exist_ok=True)

    # create logger
    logger = get_logger(args)
    logger.info(args)
    
    # seed everything
    seed_everything(args.seed)

    run(args, logger)

if __name__ == "__main__":
    main()


import os
import pickle
import guidance
import glob
import itertools

from .default_trainer import DefaultTrainer
from .utils import clean_string, _load_prompt, deduplicate
from .node import Node

class APETrainer(DefaultTrainer):
    ## simplify/customize the default training loop for APE and iterative APE
    
    def prepare_meta_prompts(self):
        # there is only initializer and proposer in APE
        
        # initializer
        init_prompt = _load_prompt(os.path.join(self.args.meta_prompts_dir, "initializer.md"))
        generation_config = "n={} temperature={} max_tokens={}".format(self.args.init_n_prompts, self.args.init_temperature, self.args.prompt_max_tokens)
        prompt = init_prompt.replace("[[GENERATION_CONFIG]]", generation_config)
        self.initializer = guidance(prompt, llm=self.llm) # guidance program for initialization

        # proposer: ~ optimizer.step()
        proposer_prompt = _load_prompt(os.path.join(self.args.meta_prompts_dir, "proposer.md"))
        generation_config = "n={} temperature={} max_tokens={}".format(self.args.n_expand, self.args.init_temperature, self.args.prompt_max_tokens) # re-using `init_temperature`
        prompt = proposer_prompt.replace("[[GENERATION_CONFIG]]", generation_config)
        self.proposer = guidance(prompt, llm=self.llm)

    def train(self, model, initial_nodes, task):
        # no gradient, just expand
        
        train_data = task.get_data_split("train") # training data
        dev_data = task.get_data_split("dev") # dev data
        
        T = self.args.train_steps # total optimization steps
        
        # initialization/loading states
        checkpoint_path = os.path.join(self.args.output_dir, "checkpoint.pkl")
        if os.path.exists(checkpoint_path) and self.args.resume:
            states, t = self.load_session(checkpoint_path)
            self.logger.info("Resuming from checkpoint {}; t={}".format(checkpoint_path, t))
        else:
            states, t = [initial_nodes], 0
            self.logger.info("Start training from scratch".format(checkpoint_path))
            
        # initialization only methods (original/non-iterative APE)
        if T == 0:
            selection_criterion = "dev"
            for i, node in enumerate(states[-1]):
                dev_results, dev_score = self.evaluate(model, node, task, dev_data, "dev")
                self.logger.info("[Node {}] Dev Score: {}".format(node.id, dev_score))
                
                node.save(self.args.output_dir)
                self.save_session(checkpoint_path, states, t)
                
        # iterative APE
        while t < T:
            
            # evaluate all prompt candidates
            for i, node in enumerate(states[-1]):
                # train_results, train_score = self.evaluate(model, node, task, train_data, "train")
                # self.logger.info("[Node {}] Train Score: {}".format(node.id, train_score))
                if self.args.do_validate:
                    dev_results, dev_score = self.evaluate(model, node, task, dev_data, "dev")
                    self.logger.info("[Node {}] Dev Score: {}".format(node.id, dev_score))
                    
                node.save(self.args.output_dir)

                # save checkpoint (save evaluation progress)
                self.save_session(checkpoint_path, states, t)
                
            # select top k prompts
            selection_criterion = "dev" if self.args.do_validate else "train"
            candidates = list(itertools.chain(*states)) if self.args.backtrack else states[-1]
            selected_nodes = sorted(candidates, key=lambda x: x.scores[selection_criterion], reverse=True)[:self.args.n_beam] # select best
            # self.logger.info("[t={}] Selected nodes: {}".format(t, str([node.id for node in selected_nodes])))
            for node in selected_nodes:
                self.logger.info("[t={}] [Node: {}] Prompt: {}".format(t, node.id, node.prompt))

            # at the last timestamp do not expand new prompt candidates
            if t == T-1:
                break
            
            # expand new prompt candidates
            new_nodes = []
            for node in selected_nodes:
                # update the prompt
                new_nodes_from_one_node = self._propose_new_nodes(node)
                new_nodes += new_nodes_from_one_node
                    
            # deduplicate
            new_nodes = deduplicate(new_nodes, prev_nodes=list(itertools.chain(*states)))
            states.append(new_nodes)

            # increment the timestamp
            t += 1
            
            # save checkpoint
            self.save_session(checkpoint_path, states, t)

        # find the best node
        best_score, best_node = -1e10, None
        for t in range(len(states)):
            for node in states[t]:
                if node.scores[selection_criterion] > best_score:
                    best_score = node.scores[selection_criterion]
                    best_node = node 

        return best_score, best_node, states
    
    def _propose_new_nodes(self, node):
        guidance.llms.OpenAI.cache.clear()
        f = self.proposer(prompt=node.prompt, max_tokens=self.args.prompt_max_tokens)
        new_nodes = []

        # directly sample n different new prompts
        for idx, item in enumerate(f["new_prompt"]):
            new_prompt = clean_string(item)
            new_node = Node(
                timestamp=node.timestamp+1,
                id=node.id+"-{}".format(node.n_child),
                parent=node.id,
                prompt=new_prompt
            )
            new_node.save(self.args.output_dir)
            new_nodes.append(new_node)
            node.n_child += 1

        return new_nodes



from .direct import DirectModel
from .zeroshotcot import ZeroshotCoTModel

MODEL2CLASS = {
    "direct": DirectModel,
    "zeroshotcot": ZeroshotCoTModel
}

def Model2Class(task):
    if task not in MODEL2CLASS:
        raise NotImplementedError
    else:
        return MODEL2CLASS[task]


import os
import logging
import pytest
from typing import Union, List

from textgrad import Variable, BlackboxLLM, TextLoss
from textgrad.optimizer import TextualGradientDescent
from textgrad.engine.vllm import ChatVLLM

logging.disable(logging.CRITICAL)
vllm_engine = ChatVLLM(model_string="meta-llama/Meta-Llama-3-8B-Instruct")

def test_import_vllm():
    assert ChatVLLM

def test_simple_forward_pass_engine():
    text = Variable("Hello", role_description="A variable")
    engine = BlackboxLLM(engine=vllm_engine)
    response = engine(text)

    assert response

def test_primitives():
    """
    Test the basic functionality of the Variable class.
    """
    x = Variable("A sntence with a typo", role_description="The input sentence", requires_grad=True)
    system_prompt = Variable("Evaluate the correctness of this sentence", role_description="The system prompt")
    loss = TextLoss(system_prompt, engine=vllm_engine)
    optimizer = TextualGradientDescent(parameters=[x], engine=vllm_engine)

    l = loss(x)
    l.backward(vllm_engine)
    optimizer.step()

    assert x.value == "A sentence with a typo"



import re
import json
import argparse
import concurrent
from tqdm import tqdm
import numpy as np
from dotenv import load_dotenv
load_dotenv(override=True)
from statistics import multimode

import textgrad as tg
from textgrad.tasks import load_instance_task


def config():
    parser = argparse.ArgumentParser(description="Optimize a prompt for a task.")
    parser.add_argument("--task", type=str, default="MMLU_machine_learning", help="The task to evaluate the model on.")
    parser.add_argument("--engine", type=str, default="gpt-4o", help="The API to use for evaluation.")
    parser.add_argument("--max_iterations", type=int, default=3, help="The maximum number of iterations of test-time updates.")
    parser.add_argument("--num_threads", type=int, default=16, help="The number of threads to use for evaluation.")
    return parser.parse_args()


class MajorityVoting:
    def __init__(self):
        pass

    def __call__(self, predictions):
        ANSWER_PATTERN_MULTICHOICE = r"(?i)Answer\s*:\s*([A-D])"
        pred_labels = []
        for pred in predictions:
            match = re.search(ANSWER_PATTERN_MULTICHOICE, pred.value)
            extracted_answer = match.group(1) if match else None
            pred_labels.append(extracted_answer)
        
        modes = multimode(pred_labels)
        return tg.Variable(f"Answer: {modes[0]}", role_description="Majority ensemble")


def get_zeroshot_answer(question):
    """Getting the zero-shot answer from an LLM without optimizing the response at test time."""
    # The system prompt is from: https://github.com/openai/simple-evals/blob/main/sampler/chat_completion_sampler.py
    STARTING_SYSTEM_PROMPT = (
    "You are ChatGPT, a large language model trained by OpenAI, based on the GPT-4 architecture."
    + "\nKnowledge cutoff: 2023-12\nCurrent date: 2024-04-01"
)
    system_prompt = tg.Variable(STARTING_SYSTEM_PROMPT, requires_grad=False, role_description="system prompt to the language model")
    model = tg.BlackboxLLM(llm_engine, system_prompt)
    response = model(tg.Variable(question, requires_grad=False, role_description="question to the language model"))
    return response

def run_test_time_training(sample):
    performance_history = []
    question, answer, test_time_objective, instance_eval_fn = sample
    zero_shot_response = get_zeroshot_answer(question)
    
    instance_var = tg.Variable(zero_shot_response.value,
                               requires_grad=True,
                               role_description="creative and precise solution and the prediction for the multiple choice question")
    
    # Evaluate the zero-shot response
    performance_history.append(int(instance_eval_fn(instance_var)))
    
    optimizer = tg.TextualGradientDescent(engine=llm_engine, 
                                          parameters=[instance_var], 
                                          constraints=["The last line of your response should be of the following format: 'Answer: $LETTER' (without quotes) where LETTER is one of ABCD."])

    predictions = []
    predictions.append(tg.Variable(
        instance_var.value,
        role_description=instance_var.role_description
        ))

    # Start test time training
    for _ in range(args.max_iterations):
        optimizer.zero_grad()
        # Compute the test time loss
        test_time_loss = test_time_objective(instance_var)
        test_time_loss.backward()
        optimizer.step()
        performance_history.append(instance_eval_fn(instance_var))
        predictions.append(tg.Variable(
        instance_var.value,
        role_description=instance_var.role_description
        ))

    ensembled_prediction = ensembler(predictions)
    performance_history.append(instance_eval_fn(ensembled_prediction))
    predictions.append(ensembled_prediction)
    return performance_history, predictions, question, answer


args = config()
llm_engine = tg.get_engine(engine_name=args.engine)
tg.set_backward_engine(llm_engine, override=True)
test_set = load_instance_task(args.task, evaluation_api=llm_engine)
ensembler = MajorityVoting()

all_solutions = {}
with concurrent.futures.ThreadPoolExecutor(max_workers=args.num_threads) as executor:
    futures = []
    for _, sample in enumerate(test_set):
        future = executor.submit(run_test_time_training, sample)
        futures.append(future)

    all_history = []
    for future in tqdm(concurrent.futures.as_completed(futures), total=len(futures), position=0):
        performance_history, predictions, question, answer = future.result()
        all_solutions[question] = {"predictions": [p.value for p in predictions], "answer": answer}
        all_history.append(performance_history)

print(np.array(all_history).mean(axis=0))
with open(f"./{args.task}_predictions.json", "w") as f:
    json.dump(all_solutions, f)


try:
    from vllm import LLM, SamplingParams
except ImportError:
    raise ImportError(
        "If you'd like to use VLLM models, please install the vllm package by running `pip install vllm` or `pip install textgrad[vllm]."
    )

import os
import platformdirs
from .base import EngineLM, CachedEngine


class ChatVLLM(EngineLM, CachedEngine):
    # Default system prompt for VLLM models
    DEFAULT_SYSTEM_PROMPT = ""

    def __init__(
        self,
        model_string="meta-llama/Meta-Llama-3-8B-Instruct",
        system_prompt=DEFAULT_SYSTEM_PROMPT,
    ):
        root = platformdirs.user_cache_dir("textgrad")
        cache_path = os.path.join(root, f"cache_vllm_{model_string}.db")
        super().__init__(cache_path=cache_path)

        self.model_string = model_string
        self.system_prompt = system_prompt
        self.client = LLM(self.model_string)
        self.tokenizer = self.client.get_tokenizer()

    def generate(
        self, prompt, system_prompt=None, temperature=0, max_tokens=2000, top_p=0.99
    ):
        sys_prompt_arg = system_prompt if system_prompt else self.system_prompt
        cache_or_none = self._check_cache(sys_prompt_arg + prompt)
        if cache_or_none is not None:
            return cache_or_none

        # The chat template ignores the system prompt;
        conversation = []
        if sys_prompt_arg:
            conversation = [{"role": "system", "content": sys_prompt_arg}]

        conversation += [{"role": "user", "content": prompt}]
        chat_str = self.tokenizer.apply_chat_template(conversation, tokenize=False)

        sampling_params = SamplingParams(
            temperature=temperature, max_tokens=max_tokens, top_p=top_p, n=1
        )

        response = self.client.generate([chat_str], sampling_params)
        response = response[0].outputs[0].text

        self._save_cache(sys_prompt_arg + prompt, response)

        return response

    def __call__(self, prompt, **kwargs):
        return self.generate(prompt, **kwargs)
